#
# Author: Guillermo Gonzalez <guillermo.gonzalez@canonical.com>
#
# Copyright 2009-2010 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
""" Base tests cases and test utilities """
from __future__ import with_statement

import dbus
from dbus.mainloop.glib import DBusGMainLoop
import contextlib
import logging
import os
import shutil
import itertools

from ubuntu_sso import DBUS_IFACE_AUTH_NAME
from ubuntu_sso.main import Login, LoginProcessor

from twisted.internet import defer
from twisted.python import failure
from unittest import TestCase
from zope.interface import implements
from oauth import oauth


@contextlib.contextmanager
def environ(env_var, new_value):
    """context manager to replace/add an environ value"""
    old_value = os.environ.get(env_var, None)
    os.environ[env_var] = new_value
    yield
    if old_value is None:
        os.environ.pop(env_var)
    else:
        os.environ[env_var] = old_value


class BaseTestCase(TestCase):
    """Base TestCase with helper methods to handle temp dir.

    This class provides:
        mktemp(name): helper to create temporary dirs
        rmtree(path): support read-only shares
        makedirs(path): support read-only shares
    """

    def mktemp(self, name='temp'):
        """Customized mktemp that accepts an optional name argument. """
        tempdir = os.path.join(self.tmpdir, name)
        if os.path.exists(tempdir):
            self.rmtree(tempdir)
        self.makedirs(tempdir)
        return tempdir

    @property
    def tmpdir(self):
        """Default tmpdir: module/class/test_method."""
        # check if we already generated the root path
        root_dir = getattr(self, '__root', None)
        if root_dir:
            return root_dir
        MAX_FILENAME = 32  # some platforms limit lengths of filenames
        base = os.path.join(self.__class__.__module__[:MAX_FILENAME],
                            self.__class__.__name__[:MAX_FILENAME],
                            self._testMethodName[:MAX_FILENAME])
        # use _trial_temp dir, it should be os.gwtcwd()
        # define the root temp dir of the testcase, pylint: disable-msg=W0201
        self.__root = os.path.join(os.getcwd(), base)
        return self.__root

    def rmtree(self, path):
        """Custom rmtree that handle ro parent(s) and childs."""
        if not os.path.exists(path):
            return
        # change perms to rw, so we can delete the temp dir
        if path != getattr(self, '__root', None):
            os.chmod(os.path.dirname(path), 0755)
        if not os.access(path, os.W_OK):
            os.chmod(path, 0755)
        # pylint: disable-msg=W0612
        for dirpath, dirs, files in os.walk(path):
            for dir in dirs:
                if not os.access(os.path.join(dirpath, dir), os.W_OK):
                    os.chmod(os.path.join(dirpath, dir), 0777)
        shutil.rmtree(path)

    def makedirs(self, path):
        """Custom makedirs that handle ro parent."""
        parent = os.path.dirname(path)
        if os.path.exists(parent):
            os.chmod(parent, 0755)
        os.makedirs(path)


class DBusTestCase(BaseTestCase):
    """ Test the DBus event handling """

    def setUp(self):
        """ Setup the infrastructure fo the test (dbus service). """
        BaseTestCase.setUp(self)
        self.loop = DBusGMainLoop(set_as_default=True)
        self.bus = dbus.bus.BusConnection(mainloop=self.loop)
        # monkeypatch busName.__del__ to avoid errors on gc
        # we take care of releasing the name in shutdown
        dbus.service.BusName.__del__ = lambda _: None
        self.bus.set_exit_on_disconnect(False)
        self.signal_receivers = set()

    def tearDown(self):
        """ Cleanup the test. """
        d = self.cleanup_signal_receivers(self.signal_receivers)
        d.addBoth(self._tearDown)
        d.addBoth(lambda _: BaseTestCase.tearDown(self))
        return d

    def _tearDown(self, *args):
        """ shutdown """
        self.dbus_iface.bus.flush()
        self.bus.flush()
        self.bus.close()

    def error_handler(self, error):
        """ default error handler for DBus calls. """
        if isinstance(error, failure.Failure):
            self.fail(error.getErrorMessage())

    def cleanup_signal_receivers(self, signal_receivers):
        """ cleanup self.signal_receivers and returns a deferred """
        deferreds = []
        for match in signal_receivers:
            d = defer.Deferred()

            def callback(*args):
                """ callback that accepts *args. """
                if not d.called:
                    d.callback(args)

            self.bus.call_async(dbus.bus.BUS_DAEMON_NAME,
                                dbus.bus.BUS_DAEMON_PATH,
                                dbus.bus.BUS_DAEMON_IFACE, 'RemoveMatch', 's',
                                (str(match),), callback, self.error_handler)
            deferreds.append(d)
        if deferreds:
            return defer.DeferredList(deferreds)
        else:
            return defer.succeed(True)


class FakeLoginProcessor(LoginProcessor):
    """Stub login processor."""

    def __init__(self, dbus_object):
        """Initialize the login processor."""
        LoginProcessor.__init__(self, dbus_object)
        self.next_login_cb = None

    def login(self, realm, consumer_key, error_handler=None,
              reply_handler=None, do_login=True):
        """Stub, call self.next_login_cb or send NewCredentials if
        self.next_login_cb isn't defined.
        """
        self.realm = str(realm)
        self.consumer_key = str(consumer_key)
        if self.next_login_cb:
            cb = self.next_login_cb[0]
            args = self.next_login_cb[1]
            self.next_login_cb = None
            return cb(*args)
        else:
            self.dbus_object.NewCredentials(realm, consumer_key)

    def clear_token(self, realm, consumer_key):
        """Stub, do nothing"""
        pass

    def next_login_with(self, callback, args=tuple()):
        """shortcircuit the next call to login and call the specified callback.
        callback is usually one of: self.got_token, self.got_no_token,
        self.got_denial or self.got_error.
        """
        self.next_login_cb = (callback, args)


class FakeLogin(Login):
    """Stub Object which listens for D-Bus OAuth requests"""

    def __init__(self, bus):
        """Initiate the object."""
        self.bus = bus
        self.busName = dbus.service.BusName(DBUS_IFACE_AUTH_NAME, bus=self.bus)
        # bypass the parent class __init__ as it has the path hardcoded
        # and we can't use '/' as the path, as we are already using it
        # for syncdaemon. pylint: disable-msg=W0233,W0231
        dbus.service.Object.__init__(self, object_path="/oauthdesktop",
                                     bus_name=self.busName)
        self.processor = FakeLoginProcessor(self)
        self.currently_authing = False

    def shutdown(self):
        """Shutdown and remove any trace from the bus"""
        self.busName.get_bus().release_name(self.busName.get_name())
        self.remove_from_connection()


class MementoHandler(logging.Handler):
    """A handler class which store logging records in a list."""

    def __init__(self, *args, **kwargs):
        """Create the instance, and add a records attribute."""
        logging.Handler.__init__(self, *args, **kwargs)
        self.records = []

    def emit(self, record):
        """Just add the record to self.records."""
        self.records.append(record)

    def check(self, level, msg):
        """Check that something is logged."""
        for rec in self.records:
            if rec.levelno == level and str(msg) in rec.message:
                return True
        return False
