# ubuntu_sso.gui - GUI for login and registration
#
# Author: Natalia Bidart <natalia.bidart@canonical.com>
#
# Copyright 2010 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.

"""User registration GUI."""

import logging
import os
import re
import tempfile

from functools import wraps

import dbus
import gettext
import gobject
import gtk
import webkit
import xdg

from dbus.mainloop.glib import DBusGMainLoop

from ubuntu_sso import DBUS_PATH, DBUS_BUS_NAME, DBUS_IFACE_USER_NAME
from ubuntu_sso.logger import setupLogging


_ = gettext.gettext
DBusGMainLoop(set_as_default=True)
logger = setupLogging('ubuntu_sso.gui')


NO_OP = lambda *args, **kwargs: None
DEFAULT_WIDTH = 30
# XXX: to be replaced by values from the theme
# see LP: #616526
HELP_TEXT_COLOR = gtk.gdk.Color("#bfbfbf")
WARNING_TEXT_COLOR = gtk.gdk.Color("red")

SIG_LOGIN_FAILED = 'login-failed'
SIG_LOGIN_SUCCEEDED = 'login-succeeded'
SIG_REGISTRATION_FAILED = 'registration-failed'
SIG_REGISTRATION_SUCCEEDED = 'registration-succeeded'
SIG_USER_CANCELATION = 'user-cancelation'

SIGNAL_ARGUMENTS = [
    (SIG_LOGIN_FAILED, ()),
    (SIG_LOGIN_SUCCEEDED, (gobject.TYPE_STRING,)),
    (SIG_REGISTRATION_FAILED, ()),
    (SIG_REGISTRATION_SUCCEEDED, (gobject.TYPE_STRING,)),
    (SIG_USER_CANCELATION, ()),
]

for sig, args in SIGNAL_ARGUMENTS:
    gobject.signal_new(sig, gtk.Window, gobject.SIGNAL_RUN_FIRST,
                       gobject.TYPE_NONE, args)


def get_data_dir():
    """Build absolute path to  the 'data' directory."""
    module = os.path.dirname(__file__)
    result = os.path.abspath(os.path.join(module, os.pardir, 'data'))
    logger.debug('get_data_file: trying to load from %r (exists? %s)',
                 result, os.path.exists(result))
    if os.path.exists(result):
        logger.info('get_data_file: returning data dir located at %r.', result)
        return result

    # no local data dir, looking within system data dirs
    data_dirs = xdg.BaseDirectory.xdg_data_dirs
    for path in data_dirs:
        result = os.path.join(path, 'ubuntu-sso-client', 'data')
        result = os.path.abspath(result)
        logger.debug('get_data_file: trying to load from %r (exists? %s)',
                     result, os.path.exists(result))
        if os.path.exists(result):
            logger.info('get_data_file: data dir located at %r.', result)
            return result
    else:
        msg = 'get_data_file: can not build a valid data path. Giving up.' \
              '__file__ is %r, data_dirs are %r'
        logger.error(msg, __file__, data_dirs)


def get_data_file(filename):
    """Build absolute path to 'filename' within the 'data' directory."""
    return os.path.join(get_data_dir(), filename)


class LabeledEntry(gtk.Entry):
    """An entry that displays the label within itself ina grey color."""

    def __init__(self, label, is_password=False, *args, **kwargs):
        self.label = label
        self.is_password = is_password
        self._user_input = None

        super(LabeledEntry, self).__init__(*args, **kwargs)

        self.set_width_chars(DEFAULT_WIDTH)
        self._set_label(self, None)
        self.set_tooltip_text(self.label)
        self.connect('focus-in-event', self._clear_text)
        self.connect('focus-out-event', self._set_label)
        self.show()

    def _clear_text(self, *args, **kwargs):
        """Clear text and restore text color."""
        self.set_text(self.get_text())

        self.modify_text(gtk.STATE_NORMAL, None)  # restore to theme's default

        if self.is_password:
            self.set_visibility(False)

        return False  # propagate the event further

    def _set_label(self, *args, **kwargs):
        """Set the proper label and proper coloring."""
        if self.get_text():
            return

        self.set_text(self.label)
        self.modify_text(gtk.STATE_NORMAL, HELP_TEXT_COLOR)

        if self.is_password:
            self.set_visibility(True)

        return False  # propagate the event further

    def get_text(self):
        """Get text only if it's not the label nor empty."""
        result = super(LabeledEntry, self).get_text()
        if result == self.label or result.isspace():
            result = ''
        return result


class UbuntuSSOClientGUI(object):
    """Ubuntu single sign on GUI."""

    CAPTCHA_SOLUTION_ENTRY = _('Type the characters above')
    CONNECT_HELP_LABEL = _('To connect this computer to %(app_name)s ' \
                           'enter your details below.')
    EMAIL1_ENTRY = _('Email address')
    EMAIL2_ENTRY = _('Re-type Email address')
    EMAIL_MISMATCH = _('The email addresses don\'t match, please double check '
                       'and try entering them again.')
    EMAIL_INVALID = _('The email must be a valid email address.')
    EMAIL_TOKEN_ENTRY = _('Enter code verification here')
    FIELD_REQUIRED = _('This field is required.')
    FORGOTTEN_PASSWORD_BUTTON = _('I\'ve forgotten my password')
    JOIN_HEADER_LABEL = _('Create %(app_name)s account')
    LOADING = _('Loading...')
    LOGIN_BUTTON_LABEL = _('Already have an account? Click here to sign in')
    LOGIN_EMAIL_ENTRY = _('Email address')
    LOGIN_HEADER_LABEL = _('Connect to %(app_name)s')
    LOGIN_PASSWORD_ENTRY = _('Password')
    NAME_ENTRY = _('Name')
    NEXT = _('Next')
    ONE_MOMENT_PLEASE = _('One moment please...')
    PASSWORD_CHANGED = _('Your password was successfully changed.')
    PASSWORD1_ENTRY = RESET_PASSWORD1_ENTRY = _('Password')
    PASSWORD2_ENTRY = RESET_PASSWORD2_ENTRY = _('Re-type Password')
    PASSWORD_HELP = _('The password must have a minimum of 8 characters and '
                      'include one uppercase character and one number.')
    PASSWORD_MISMATCH = _('The passwords don\'t match, please double check '
                          'and try entering them again.')
    PASSWORD_TOO_WEAK = _('The password is too weak.')
    REQUEST_PASSWORD_TOKEN_LABEL = _('To reset your password enter your ' \
                                     'email address below:')
    RESET_PASSWORD = _('Reset Password')
    RESET_CODE_ENTRY = _('Reset code')
    RESET_EMAIL_ENTRY = _('Email address')
    SET_NEW_PASSWORD_LABEL = _('A password reset code has been sent to ' \
                               '%(email)s.\nPlease enter the code below ' \
                               'along with your new password.')
    SUCCESS = _('The process finished successfully. Congratulations!')
    TC = _('Terms & Conditions')
    TC_NOT_ACCEPTED = _('Agreeing to the Ubuntu One Terms & Conditions is '
                        'required to subscribe.')
    UNKNOWN_ERROR = _('There was an error when trying to complete the ' \
                      'process. Please check the information and try again.')
    VERIFY_EMAIL_LABEL = _("""Enter verification code.

A verification code has just been sent to your email address.
Please enter your code from the email. An example is shown below.""")
    YES_TO_TC = _('I agree with the %(app_name)s ')  # Terms&Conditions button
    YES_TO_UPDATES = _('Yes! Email me %(app_name)s tips and updates.')

    def __init__(self, app_name, tc_uri, help_text,
                 window_id=0, login_only=False, close_callback=None):
        """Create the GUI and initialize widgets."""
        gtk.link_button_set_uri_hook(NO_OP)

        self._captcha_filename = tempfile.mktemp()
        self._captcha_id = None
        self._signals_receivers = {}

        self.app_name = app_name
        self.app_label = '<b>%s</b>' % self.app_name
        self.tc_uri = tc_uri
        self.help_text = help_text
        self.close_callback = close_callback

        ui_filename = get_data_file('ui.glade')
        builder = gtk.Builder()
        builder.add_from_file(ui_filename)
        builder.connect_signals(self)

        # XXX: every button should have tests for 'activate' signal
        # see LP: #616421

        self.widgets = []
        self.warnings = []
        self.cancels = []
        for obj in builder.get_objects():
            name = getattr(obj, 'name', None)
            if name is None and isinstance(obj, gtk.Buildable):
                # work around bug lp:507739
                name = gtk.Buildable.get_name(obj)
            if name is None:
                logging.warn("%s has no name (??)", obj)
            else:
                self.widgets.append(name)
                setattr(self, name, obj)
                if 'warning' in name:
                    self.warnings.append(obj)
                    obj.hide()
                if 'cancel_button' in name:
                    obj.connect('clicked', self.on_close_clicked)
                    self.cancels.append(obj)

        self.entries = ('name_entry', 'email1_entry', 'email2_entry',
                        'password1_entry', 'password2_entry',
                        'captcha_solution_entry', 'email_token_entry',
                        'login_email_entry', 'login_password_entry',
                        'reset_email_entry',
                        'reset_code_entry',
                        'reset_password1_entry', 'reset_password2_entry')

        for name in self.entries:
            label = getattr(self, name.upper())
            is_password = 'password' in name
            entry = LabeledEntry(label=label, is_password=is_password)
            setattr(self, name, entry)
            assert getattr(self, name) is not None

        self.window.set_icon_name('ubuntu-logo')

        self.bus = dbus.SessionBus()
        self.bus.add_signal_receiver = self._log(self.bus.add_signal_receiver)
        obj = self.bus.get_object(bus_name=DBUS_BUS_NAME,
                                  object_path=DBUS_PATH,
                                  follow_name_owner_changes=True)
        self.iface_name = DBUS_IFACE_USER_NAME
        self.backend = dbus.Interface(object=obj,
                                      dbus_interface=self.iface_name)
        logger.debug('UbuntuSSOClientGUI: backend created: %r', self.backend)

        self.pages = (self.enter_details_vbox, self.processing_vbox,
                      self.verify_email_vbox, self.success_vbox,
                      self.tc_browser_vbox, self.login_vbox,
                      self.request_password_token_vbox,
                      self.set_new_password_vbox)

        self._append_page(self._build_processing_page())
        self._append_page(self._build_success_page())
        self._append_page(self._build_login_page())
        self._append_page(self._build_request_password_token_page())
        self._append_page(self._build_set_new_password_page())

        if not login_only:
            self._append_page(self._build_enter_details_page())
            self._append_page(self._build_tc_page())
            self._append_page(self._build_verify_email_page())
            self.login_button.grab_focus()
            self._set_current_page(self.enter_details_vbox)
        else:
            self.login_back_button.hide()
            self.login_ok_button.grab_focus()
            self._set_current_page(self.login_vbox)

        self._setup_signals()
        self._gtk_signal_log = []

        if window_id != 0:
            # be as robust as possible:
            # if the window_id is not "good", set_transient_for will fail
            # awfully, and we don't want that: if the window_id is bad we can
            # still do everything as a standalone window. Also,
            # window_foreign_new may return None breaking set_transient_for.
            try:
                r = gtk.gdk.window_foreign_new(window_id)
                self.window.realize()
                self.window.window.set_transient_for(r)
            except:
                msg = 'UbuntuSSOClientGUI: failed set_transient_for win id %r'
                logger.exception(msg, window_id)
        self.window.show()

    def _setup_signals(self):
        """Bind signals to callbacks to be able to test the pages."""
        self._signals = [
            ('CaptchaGenerated', self.on_captcha_generated),
            ('CaptchaGenerationError', self.on_captcha_generation_error),
            ('UserRegistered', self.on_user_registered),
            ('UserRegistrationError', self.on_user_registration_error),
            ('EmailValidated', self.on_email_validated),
            ('EmailValidationError', self.on_email_validation_error),
            ('LoggedIn', self.on_logged_in),
            ('LoginError', self.on_login_error),
            ('PasswordResetTokenSent', self.on_password_reset_token_sent),
            ('PasswordResetError', self.on_password_reset_error),
            ('PasswordChanged', self.on_password_changed),
            ('PasswordChangeError', self.on_password_change_error),
        ]
        for signal, method in self._signals:
            self.bus.add_signal_receiver(method, signal_name=signal,
                                         dbus_interface=self.iface_name)

    def _debug(self, *args, **kwargs):
        """Do some debugging."""
        print args, kwargs

    def _log(self, f):
        """Keep track of added signals to the internal bus."""

        @wraps(f)
        def inner(method, signal_name, dbus_interface):
            """Do the deed."""
            actual = self._signals_receivers.get((dbus_interface, signal_name))
            assert actual is None
            f(method, signal_name, dbus_interface)
            self._signals_receivers[(dbus_interface, signal_name)] = method

        return inner

    def _add_spinner_to_container(self, container, legend=None):
        """Add a spinner to 'container'."""
        spinner = gtk.Spinner()
        spinner.start()

        label = gtk.Label()
        if legend:
            label.set_text(legend)
        else:
            label.set_text(self.LOADING)

        hbox = gtk.HBox(spacing=5)
        hbox.pack_start(spinner, expand=False)
        hbox.pack_start(label, expand=False)

        alignment = gtk.Alignment(xalign=0.5, yalign=0.5)
        alignment.add(hbox)
        alignment.show_all()

        container.add(alignment)

    # helpers

    def _set_warning_message(self, widget, message):
        """Set 'message' as text for 'widget'."""
        widget.set_text(message)
        widget.modify_fg(gtk.STATE_NORMAL, WARNING_TEXT_COLOR)
        widget.show()

    def _clear_warnings(self):
        """Clear all warning messages."""
        for widget in self.warnings:
            widget.set_text('')
            widget.hide()

    def _non_empty_input(self, widget):
        """Return weather widget has non empty content."""
        text = widget.get_text()
        return bool(text and not text.isspace())

    # build pages

    def _append_page(self, page):
        """Append 'page' to the 'window'."""
        self.window.get_children()[0].pack_start(page)

    def _set_header(self, header):
        """Set 'header' as the window title and header."""
        markup = '<span size="xx-large">%s</span>'
        self.header_label.set_markup(markup % header)
        self.window.set_title(self.header_label.get_text())  # avoid markup

    def _set_current_page(self, current_page, warning_text=None):
        """Hide all the pages but 'current_page'."""
        if hasattr(current_page, 'header'):
            self._set_header(current_page.header)

        if hasattr(current_page, 'help_text'):
            self.help_label.set_markup(current_page.help_text)

        if warning_text is not None:
            self._set_warning_message(self.warning_label, warning_text)
        else:
            self.warning_label.hide()

        for page in self.pages:
            if page is current_page:
                page.show()
            else:
                page.hide()

    def _set_captcha_loading(self):
        """Present a spinner to the user while the captcha is downloaded."""
        self.captcha_image.hide()
        self._add_spinner_to_container(self.captcha_loading)
        white = gtk.gdk.Color('white')
        self.captcha_loading.modify_bg(gtk.STATE_NORMAL, white)
        self.captcha_loading.show_all()
        self.join_ok_button.set_sensitive(False)

    def _set_captcha_image(self):
        """Present a captcha image to the user to be resolved."""
        self.captcha_loading.hide()
        self.join_ok_button.set_sensitive(True)
        self.captcha_image.set_from_file(self._captcha_filename)
        self.captcha_image.show()

    def _build_enter_details_page(self):
        """Build the enter details page."""
        d = {'app_name': self.app_label}
        self.enter_details_vbox.header = self.JOIN_HEADER_LABEL % d
        self.enter_details_vbox.help_text = self.help_text

        self.enter_details_vbox.pack_start(self.name_entry, expand=False)
        self.enter_details_vbox.reorder_child(self.name_entry, 0)
        self.captcha_solution_vbox.pack_start(self.captcha_solution_entry,
                                              expand=False)
        self.captcha_solution_vbox.reorder_child(self.captcha_solution_entry,
                                                 0)

        self.emails_hbox.pack_start(self.email1_entry)
        self.emails_hbox.pack_start(self.email2_entry)

        self.passwords_hbox.pack_start(self.password1_entry)
        self.passwords_hbox.pack_start(self.password2_entry)
        help_msg = '<small>%s</small>' % self.PASSWORD_HELP
        self.password_help_label.set_markup(help_msg)

        if not os.path.exists(self._captcha_filename):
            self._set_captcha_loading()
        else:
            self._set_captcha_image()

        msg = self.YES_TO_UPDATES % {'app_name': self.app_name}
        self.yes_to_updates_checkbutton.set_label(msg)
        msg = self.YES_TO_TC % {'app_name': self.app_name}
        self.yes_to_tc_checkbutton.set_label(msg)
        self.tc_button.set_label(self.TC)
        self.login_button.set_label(self.LOGIN_BUTTON_LABEL)

        logger.info('Calling generate_captcha with filename path at %r',
                    self._captcha_filename)
        f = self.backend.generate_captcha
        f(self._captcha_filename,
          reply_handler=NO_OP, error_handler=NO_OP)

        return self.enter_details_vbox

    def _build_tc_page(self):
        """Build the Terms & Conditions page."""
        return self.tc_browser_vbox

    def _build_processing_page(self):
        """Build the processing page with a spinner."""
        self._add_spinner_to_container(self.processing_vbox,
                                       legend=self.ONE_MOMENT_PLEASE)
        return self.processing_vbox

    def _build_verify_email_page(self):
        """Build the verify email page."""
        self.verify_email_vbox.help_text = self.VERIFY_EMAIL_LABEL
        self.verify_email_vbox.pack_start(self.email_token_entry, expand=False)
        self.verify_email_vbox.reorder_child(self.email_token_entry, 1)

        fname = get_data_file('email.png')
        self._email_example_pixbuf = gtk.gdk.pixbuf_new_from_file(fname)
        self.verify_email_image.set_from_pixbuf(self._email_example_pixbuf)

        return self.verify_email_vbox

    def _build_success_page(self):
        """Build the success page."""
        #self.success_vbox.help_text = ''
        self.success_label.set_markup('<span size="x-large">%s</span>' %
                                      self.SUCCESS)
        return self.success_vbox

    def _build_login_page(self):
        """Build the login page."""
        d = {'app_name': self.app_label}
        self.login_vbox.header = self.LOGIN_HEADER_LABEL % d
        self.login_vbox.help_text = self.CONNECT_HELP_LABEL % d

        self.login_details_vbox.pack_start(self.login_email_entry)
        self.login_details_vbox.reorder_child(self.login_email_entry, 0)
        self.login_details_vbox.pack_start(self.login_password_entry)
        self.login_details_vbox.reorder_child(self.login_password_entry, 2)

        msg = self.FORGOTTEN_PASSWORD_BUTTON
        self.forgotten_password_button.set_label(msg)
        self.login_ok_button.grab_focus()

        return self.login_vbox

    def _build_request_password_token_page(self):
        """Build the login page."""
        self.request_password_token_vbox.header = self.RESET_PASSWORD
        text = self.REQUEST_PASSWORD_TOKEN_LABEL
        self.request_password_token_vbox.help_text = text

        entry = self.reset_email_entry
        self.request_password_token_details_vbox.pack_start(entry)
        cb = self.on_reset_email_entry_changed
        self.reset_email_entry.connect('changed', cb)
        self.request_password_token_ok_button.set_label(self.NEXT)
        self.request_password_token_ok_button.set_sensitive(False)

        return self.request_password_token_vbox

    def _build_set_new_password_page(self):
        """Build the login page."""
        self.set_new_password_vbox.header = self.RESET_PASSWORD
        self.set_new_password_vbox.help_text = self.SET_NEW_PASSWORD_LABEL

        for entry in (self.reset_code_entry,
                      self.reset_password1_entry,
                      self.reset_password2_entry):
            self.set_new_password_details_vbox.pack_start(entry)

        cb = self.on_set_new_password_entries_changed
        self.reset_code_entry.connect('changed', cb)
        self.reset_password1_entry.connect('changed', cb)
        self.reset_password2_entry.connect('changed', cb)

        self.set_new_password_ok_button.set_label(self.RESET_PASSWORD)
        self.set_new_password_ok_button.set_sensitive(False)

        return self.set_new_password_vbox

    # GTK callbacks

    def run(self):
        """Run the application."""
        gtk.main()

    def connect(self, signal_name, handler, *args, **kwargs):
        """Connect 'signal_name' with 'handler'."""
        logger.debug('connect: signal %r, handler %r, args  %r, kwargs, %r',
                     signal_name, handler, args, kwargs)
        self.window.connect(signal_name, handler, *args, **kwargs)

    def on_close_clicked(self, *args, **kwargs):
        """Call self.close_callback if defined."""
        if os.path.exists(self._captcha_filename):
            os.remove(self._captcha_filename)

        # remove the signals from DBus
        remove = self.bus.remove_signal_receiver
        for (iface, signal) in self._signals_receivers.keys():
            method = self._signals_receivers.pop((iface, signal))
            remove(method, signal_name=signal, dbus_interface=iface)

        # destroy main window
        if self.window is not None:
            self.window.hide()

        if len(args) > 0 and args[0] in self.cancels:
            self.window.emit(SIG_USER_CANCELATION)
        elif len(self._gtk_signal_log) > 0:
            signal = self._gtk_signal_log[-1][0]
            args = self._gtk_signal_log[-1][1:]
            self.window.emit(signal, *args)

        # call user defined callback
        if self.close_callback is not None:
            logger.info('Calling custom close_callback %r with params %r, %r',
                        self.close_callback, args, kwargs)
            self.close_callback(*args, **kwargs)

    def on_sign_in_button_clicked(self, *args, **kwargs):
        """User wants to sign in, present the Login page."""
        self._set_current_page(self.login_vbox)

    def on_join_ok_button_clicked(self, *args, **kwargs):
        """Submit info for processing, present the processing vbox."""
        self._clear_warnings()

        error = False

        name = self.name_entry.get_text()
        if not name:
            self._set_warning_message(self.name_warning_label,
                                      self.FIELD_REQUIRED)
            error |= True

        # check email
        email1 = self.email1_entry.get_text()
        email2 = self.email2_entry.get_text()
        if email1 != email2:
            self._set_warning_message(self.email_warning_label,
                                      self.EMAIL_MISMATCH)
            error |= True

        if '@' not in email1:
            self._set_warning_message(self.email_warning_label,
                                      self.EMAIL_INVALID)
            error |= True

        if not email1:
            self._set_warning_message(self.email_warning_label,
                                      self.FIELD_REQUIRED)
            error |= True

        # check password
        password1 = self.password1_entry.get_text()
        password2 = self.password2_entry.get_text()
        if password1 != password2:
            self._set_warning_message(self.password_warning_label,
                                      self.PASSWORD_MISMATCH)
            error |= True

        if (len(password1) < 8 or
            re.search('[A-Z]', password1) is None or
            re.search('\d+', password1) is None):
            self._set_warning_message(self.password_warning_label,
                                      self.PASSWORD_TOO_WEAK)
            error |= True

        # check T&C
        if not self.yes_to_tc_checkbutton.get_active():
            self._set_warning_message(self.tc_warning_label,
                                      self.TC_NOT_ACCEPTED)
            error |= True

        captcha_solution = self.captcha_solution_entry.get_text()
        if not captcha_solution:
            self._set_warning_message(self.captcha_solution_warning_label,
                                      self.FIELD_REQUIRED)
            error |= True

        if error:
            return

        self._clear_warnings()
        self._set_current_page(self.processing_vbox)

        logger.info('Calling register_user with email %r, password <hidden>,' \
                    ' captcha_id %r and captcha_solution %r.', email1,
                    self._captcha_id, captcha_solution)
        f = self.backend.register_user
        f(email1, password1, self._captcha_id, captcha_solution,
          reply_handler=NO_OP, error_handler=NO_OP)

    def on_tc_button_clicked(self, *args, **kwargs):
        """T & C button was clicked, show the browser with them."""
        self._set_current_page(self.tc_browser_vbox)

    def on_tc_back_button_clicked(self, *args, **kwargs):
        """T & C 'back' button was clicked, return to the previous page."""
        self._set_current_page(self.enter_details_vbox)

    def on_verify_token_button_clicked(self, *args, **kwargs):
        """The user entered the email token, let's verify!"""
        email = self.email1_entry.get_text()
        password = self.password1_entry.get_text()
        email_token = self.email_token_entry.get_text()

        logger.info('Calling validate_email with email %r, password <hidden>' \
                    ', app_name %r and email_token %r.', email,
                    self.app_name, email_token)
        f = self.backend.validate_email
        f(email, password, self.app_name, email_token,
          reply_handler=NO_OP, error_handler=NO_OP)

        self._set_current_page(self.processing_vbox)

    def on_login_connect_button_clicked(self, *args, **kwargs):
        """User wants to connect!"""
        self._clear_warnings()

        error = False
        email = self.login_email_entry.get_text()

        if '@' not in email:
            self._set_warning_message(self.login_email_warning_label,
                                      self.EMAIL_INVALID)
            error |= True

        if not email:
            self._set_warning_message(self.login_email_warning_label,
                                      self.FIELD_REQUIRED)
            error |= True

        password = self.login_password_entry.get_text()

        if not password:
            self._set_warning_message(self.login_password_warning_label,
                                      self.FIELD_REQUIRED)
            error |= True

        if error:
            return

        self._clear_warnings()

        f = self.backend.login
        f(email, password, self.app_name,
          reply_handler=NO_OP, error_handler=NO_OP)

        self._set_current_page(self.processing_vbox)

    def on_login_back_button_clicked(self, *args, **kwargs):
        """User wants to go to the previous page."""
        self._set_current_page(self.enter_details_vbox)

    def on_forgotten_password_button_clicked(self, *args, **kwargs):
        """User wants to reset the password."""
        self._set_current_page(self.request_password_token_vbox)

    def on_request_password_token_ok_button_clicked(self, *args, **kwargs):
        """User entered the email address to reset the password."""
        email = self.reset_email_entry.get_text()

        logger.info('Calling request_password_reset_token with %r.', email)
        f = self.backend.request_password_reset_token
        f(email, reply_handler=NO_OP, error_handler=NO_OP)

        self._set_current_page(self.processing_vbox)

    def on_request_password_token_back_button_clicked(self, *args, **kwargs):
        """User wants to go to the previous page."""
        self._set_current_page(self.login_vbox)

    def on_reset_email_entry_changed(self, widget, *args, **kwargs):
        """User is changing the 'widget' entry in the reset email page."""
        sensitive = self._non_empty_input(widget)
        self.request_password_token_ok_button.set_sensitive(sensitive)

    def on_set_new_password_entries_changed(self, *args, **kwargs):
        """User is changing the 'widget' entry in the reset password page."""
        sensitive = True
        for entry in (self.reset_code_entry,
                      self.reset_password1_entry,
                      self.reset_password2_entry):
            sensitive &= self._non_empty_input(entry)
        self.set_new_password_ok_button.set_sensitive(sensitive)

    def on_set_new_password_ok_button_clicked(self, *args, **kwargs):
        """User entered reset code and new passwords."""
        email = self.reset_email_entry.get_text()
        token = self.reset_code_entry.get_text()
        # XXX: validate passwords!
        # see LP: #616528
        password1 = self.reset_password1_entry.get_text()
        #password2 = self.reset_password2_entry.get_text()

        logger.info('Calling set_new_password with email %r, token %r and ' \
                    'new password: <hidden>.', email, token)
        f = self.backend.set_new_password
        f(email, token, password1, reply_handler=NO_OP, error_handler=NO_OP)

        self._set_current_page(self.processing_vbox)

    def on_tc_browser_vbox_show(self, *args, **kwargs):
        """The T&C page is being shown."""
        browser = webkit.WebView()
        browser.open(self.tc_uri)
        browser.show()
        self.tc_browser_window.add(browser)

    def on_tc_browser_vbox_hide(self, *args, **kwargs):
        """The T&C page is no longer being shown."""
        children = self.tc_browser_window.get_children()
        if len(children) > 0:
            browser = children[0]
            self.tc_browser_window.remove(browser)
            browser.destroy()
            del browser

    # backend callbacks

    def on_captcha_generated(self, captcha_id, *args, **kwargs):
        """Captcha image has been generated and is available to be shown."""
        assert captcha_id is not None
        self._captcha_id = captcha_id
        self._set_captcha_image()

    def on_captcha_generation_error(self, *args, **kwargs):
        """Captcha image generation failed."""
        logger.warning('on_captcha_generation_error: args %r, kwargs %r',
                       args, kwargs)

    def on_user_registered(self, *args, **kwargs):
        """Registration can go on, user needs to verify email."""
        logger.info('on_user_registered: user was successfully registered! ' \
                    'args %r, kwargs %r.', args, kwargs)
        self._set_current_page(self.verify_email_vbox)

    def on_user_registration_error(self, *args, **kwargs):
        """Captcha image generation failed."""
        logger.warning('on_user_registration_error: args %r, kwargs %r',
                       args, kwargs)
        self._set_current_page(self.enter_details_vbox,
                               warning_text=self.UNKNOWN_ERROR)
        self._gtk_signal_log.append((SIG_REGISTRATION_FAILED,))

    def on_email_validated(self, app_name, *args, **kwargs):
        """User email was successfully verified."""
        logger.info('on_email_validated: email was successfully validated! ' \
                    'app_name %r, args %r, kwargs %r.', app_name, args, kwargs)
        self._set_current_page(self.success_vbox)
        self._gtk_signal_log.append((SIG_REGISTRATION_SUCCEEDED, app_name))

    def on_email_validation_error(self, *args, **kwargs):
        """User email validation failed."""
        logger.warning('on_email_validation_error: args %r, kwargs %r',
                       args, kwargs)
        self._set_current_page(self.verify_email_vbox,
                               warning_text=self.UNKNOWN_ERROR)
        self._gtk_signal_log.append((SIG_REGISTRATION_FAILED,))

    def on_logged_in(self, app_name, *args, **kwargs):
        """User was successfully logged in."""
        logger.info('on_logged_in: user was successfully logged in! '
                    'args %r, kwargs %r', args, kwargs)
        self._set_current_page(self.success_vbox)
        self._gtk_signal_log.append((SIG_LOGIN_SUCCEEDED, app_name))

    def on_login_error(self, *args, **kwargs):
        """User was not successfully logged in."""
        logger.warning('on_login_error: args %r, kwargs %r',
                       args, kwargs)
        self._set_current_page(self.login_vbox,
                               warning_text=self.UNKNOWN_ERROR)
        self._gtk_signal_log.append((SIG_LOGIN_FAILED,))

    def on_password_reset_token_sent(self, email, *args, **kwargs):
        """Password reset token was successfully sent."""
        email = self.reset_email_entry.get_text()
        msg = self.SET_NEW_PASSWORD_LABEL % {'email': email}
        self.set_new_password_vbox.help_text = msg
        self._set_current_page(self.set_new_password_vbox)

    def on_password_reset_error(self, *args, **kwargs):
        """Password reset failed."""
        logger.warning('on_password_reset_error: args %r, kwargs %r',
                       args, kwargs)
        self._set_current_page(self.login_vbox,
                               warning_text=self.UNKNOWN_ERROR)

    def on_password_changed(self, email, *args, **kwargs):
        """Password was successfully changed."""
        self._set_current_page(self.login_vbox,
                               warning_text=self.PASSWORD_CHANGED)

    def on_password_change_error(self, *args, **kwargs):
        """Password reset failed."""
        logger.warning('on_password_change_error: args %r, kwargs %r',
                       args, kwargs)
        self._set_current_page(self.request_password_token_vbox,
                               warning_text=self.UNKNOWN_ERROR)
