# Copyright (C) 2010 Canonical
#
# Authors:
#  Andrew Higginson
#  Alejandro J. Cura <alecu@canonical.com>
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; version 3.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Handle keys in the gnome kerying."""

import gnomekeyring

from urllib import urlencode
from urlparse import parse_qsl


class Keyring(object):

    KEYRING_NAME = "login"

    def __init__(self, app_name):
        if not gnomekeyring.is_available():
            raise gnomekeyring.NoKeyringDaemonError
        self.app_name = app_name

    def _create_keyring(self, name):
        """Creates a keyring, if it already exists, do nothing."""
        keyring_names = gnomekeyring.list_keyring_names_sync()
        if not name in keyring_names:
            gnomekeyring.create_sync(name)

    def _get_item_id_from_name(self, sync, name):
        """Return the ID for a named item."""
        for item_id in gnomekeyring.list_item_ids_sync(sync):
            item_info = gnomekeyring.item_get_info_sync(sync, item_id)
            display_name = item_info.get_display_name()
            if display_name == name:
                return item_id

    def _get_item_info_from_name(self, sync, name):
        """Return the ID for a named item."""
        for item_id in gnomekeyring.list_item_ids_sync(sync):
            item_info = gnomekeyring.item_get_info_sync(sync, item_id)
            display_name = item_info.get_display_name()
            if display_name == name:
                return item_info

    def set_ubuntusso_attr(self, cred):
        """Set the credentials of the Ubuntu SSO item."""
        # Creates the secret from the credentials
        secret = urlencode(cred)

        # Create the keyring
        self._create_keyring(self.KEYRING_NAME)

        # No need to delete the item if it already exists

        # A set of attributes for this credentials
        attr = {"key-type": "Ubuntu SSO credentials",
                "token-name": cred["name"].encode("utf8")}

        # Add our SSO credentials to the keyring
        gnomekeyring.item_create_sync(self.KEYRING_NAME,
                gnomekeyring.ITEM_GENERIC_SECRET, self.app_name, attr,
                secret, True)

    def get_ubuntusso_attr(self):
        """Return the secret of the SSO item in a dictionary."""
        # If we have no attributes, return None
        exist = self._get_item_info_from_name(self.KEYRING_NAME, self.app_name)
        if exist is not None:
            secret = self._get_item_info_from_name(self.KEYRING_NAME,
                                                   self.app_name).get_secret()
            return dict(parse_qsl(secret))

    def delete_ubuntusso_attr(self):
        """Delete a set of credentials from the keyring."""
        item_id = self._get_item_id_from_name(self.KEYRING_NAME,
                                              self.app_name)
        if item_id is not None:
            gnomekeyring.item_delete_sync(self.KEYRING_NAME, item_id)


if __name__ == "__main__":
    kr1 = Keyring("Test key 1")
    kr2 = Keyring("Test key 2")

    kr1.delete_ubuntusso_attr()
    kr2.delete_ubuntusso_attr()

    d1 = {"name": "test-1", "ha": "hehddddeff", "hi": "hggehes", "ho": "he"}
    d2 = {"name": "test-2", "hi": "ho", "let's": "go"}

    kr1.set_ubuntusso_attr(d1)
    kr2.set_ubuntusso_attr(d2)

    r1 = kr1.get_ubuntusso_attr()
    r2 = kr2.get_ubuntusso_attr()

    assert r1 == d1
    assert r2 == d2
