#!/usr/bin/python

# ubuntu_sso.main - main login handling interface
#
# Author: Stuart Langridge <stuart.langridge@canonical.com>
# Author: Natalia Bidart <natalia.bidart@canonical.com>
# Author: Alejandro J. Cura <alecu@canonical.com>
#
# Copyright 2009 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
"""OAuth login handler.

A command-line utility which accepts requests for OAuth login over D-Bus,
handles the OAuth process (including adding the OAuth access token to the
gnome keyring), and then alerts the calling app (and others) with a D-Bus
signal so they can retrieve the new token.
"""

import re
import socket
import time
import threading
import urllib
import urllib2
import urlparse

import dbus.service
import gobject

from dbus.mainloop.glib import DBusGMainLoop
from lazr.restfulclient.authorize import BasicHttpAuthorizer
from lazr.restfulclient.authorize.oauth import OAuthAuthorizer
from lazr.restfulclient.errors import HTTPError
from lazr.restfulclient.resource import ServiceRoot
from oauth.oauth import OAuthToken

from keyring import Keyring
from ubuntu_sso import (DBUS_IFACE_AUTH_NAME, DBUS_IFACE_USER_NAME,
    DBUS_IFACE_CRED_NAME, DBUS_CRED_PATH, DBUS_BUS_NAME, gui)
from ubuntu_sso.config import get_config
from ubuntu_sso.logger import setupLogging
logger = setupLogging("ubuntu_sso.main")

DBusGMainLoop(set_as_default=True)
# Disable the invalid name warning, as we have a lot of DBus style names
# pylint: disable-msg=C0103

OLD_KEY_NAME = "UbuntuOne token for https://ubuntuone.com"
U1_APP_NAME = "Ubuntu One"


class NoDefaultConfigError(Exception):
    """No default section in configuration file"""


class BadRealmError(Exception):
    """Realm must be a URL."""


class InvalidEmailError(Exception):
    """The email is not valid."""


class InvalidPasswordError(Exception):
    """The password is not valid.

    Must provide at least 8 characters, one upper case, one number.
    """


class RegistrationError(Exception):
    """The registration failed."""


class AuthenticationError(Exception):
    """The authentication failed."""


class EmailTokenError(Exception):
    """The email token is not valid."""


class ResetPasswordTokenError(Exception):
    """The token for password reset could not be generated."""


class NewPasswordError(Exception):
    """The new password could not be set."""


def keyring_store_credentials(app_name, credentials):
    """Store the credentials in the keyring."""
    Keyring(app_name).set_ubuntusso_attr(credentials)


def keyring_get_credentials(app_name):
    """Get the credentials from the keyring or None if not there."""
    creds = Keyring(app_name).get_ubuntusso_attr()
    if creds is None and app_name == U1_APP_NAME:
        # No new creds, try to get old credentials
        old_creds = Keyring(OLD_KEY_NAME).get_ubuntusso_attr()
        if old_creds is not None:
            # Old creds found, build a new credentials dict with them
            creds = {
                'consumer_key': "ubuntuone",
                'consumer_secret': "hammertime",
                'name': U1_APP_NAME,
                'token': old_creds["oauth_token"],
                'token_secret': old_creds["oauth_token_secret"],
            }
    return creds


def get_token_name(app_name):
    """Build the token name."""
    quoted_app_name = urllib.quote(app_name)
    computer_name = socket.gethostname()
    quoted_computer_name = urllib.quote(computer_name)
    return "%s - %s" % (quoted_app_name, quoted_computer_name)


class SSOLoginProcessor(object):
    """Login and register users using the Ubuntu Single Sign On service."""

    def __init__(self, sso_service_class=None):
        """Create a new SSO login processor."""
        if sso_service_class is None:
            self.sso_service_class = ServiceRoot
        else:
            self.sso_service_class = sso_service_class

        self.service_url = 'https://login.ubuntu.com/api/1.0'

    def _valid_email(self, email):
        """Validate the given email."""
        # XXX: to be improved as per django validation
        return email is not None and '@' in email

    def _valid_password(self, password):
        """Validate the given password."""
        res = (len(password) > 7 and  # at least 8 characters
               re.search('[A-Z]', password) and  # one upper case
               re.search('\d+', password))  # one number
        return res

    def generate_captcha(self, filename):
        """Generate a captcha using the SSO service."""
        logger.debug('generate_captcha: requesting captcha, filename: %r',
                     filename)
        sso_service = self.sso_service_class(None, self.service_url)
        captcha = sso_service.captchas.new()

        # download captcha and save to 'filename'
        logger.debug('generate_captcha: server answered: %r', captcha)
        try:
            res = urllib2.urlopen(captcha['image_url'])
            with open(filename, 'wb') as f:
                f.write(res.read())
        except:
            msg = 'generate_captcha crashed while downloading the image.'
            logger.exception(msg)
            raise

        return captcha['captcha_id']

    def register_user(self, email, password, captcha_id, captcha_solution):
        """Register a new user with 'email' and 'password'."""
        logger.debug('register_user: email: %r password: <hidden>, '
                     'captcha_id: %r, captcha_solution: %r',
                     email, captcha_id, captcha_solution)
        sso_service = self.sso_service_class(None, self.service_url)
        if not self._valid_email(email):
            logger.error('register_user: InvalidEmailError for email: %r',
                         email)
            raise InvalidEmailError()
        if not self._valid_password(password):
            logger.error('register_user: InvalidPasswordError')
            raise InvalidPasswordError()

        result = sso_service.registrations.register(
                    email=email, password=password, captcha_id=captcha_id,
                    captcha_solution=captcha_solution)
        logger.info('register_user: email: %r result: %r', email, result)

        if result['status'] == 'error':
            raise RegistrationError(result['errors'])
        elif result['status'] != 'ok':
            raise RegistrationError('Received unknown status: %s' % result)
        else:
            return True

    def login(self, email, password, app_name):
        """Login a user with 'email' and 'password'."""
        logger.debug('login: email: %r password: <hidden>, app_name: %r',
                     email, app_name)
        basic = BasicHttpAuthorizer(email, password)
        sso_service = self.sso_service_class(basic, self.service_url)
        service = sso_service.authentications.authenticate

        token_name = get_token_name(app_name)
        try:
            credentials = service(token_name=token_name)
        except HTTPError:
            logger.exception('login failed with:')
            raise AuthenticationError()

        logger.debug('login: authentication successful! consumer_key: %r, ' \
                     'token_name: %r', credentials['consumer_key'], token_name)
        keyring_store_credentials(app_name, credentials)
        return credentials

    def validate_email(self, email, password, app_name, email_token):
        """Validate an email token for user with 'email' and 'password'."""
        logger.debug('validate_email: email: %r password: <hidden>, '
                     'app_name: %r, email_token: %r',
                     email, app_name, email_token)
        token = self.login(email=email, password=password,
                           app_name=app_name)

        oauth_token = OAuthToken(token['token'], token['token_secret'])
        authorizer = OAuthAuthorizer(token['consumer_key'],
                                     token['consumer_secret'],
                                     oauth_token)
        sso_service = self.sso_service_class(authorizer, self.service_url)
        result = sso_service.accounts.validate_email(email_token=email_token)
        logger.info('validate_email: email: %r result: %r', email, result)
        if 'errors' in result:
            raise EmailTokenError(result['errors'])
        elif 'email' in result:
            return True
        else:
            raise EmailTokenError('Received invalid reply: %s' % result)

    def request_password_reset_token(self, email):
        """Request a token to reset the password for the account 'email'."""
        sso_service = self.sso_service_class(None, self.service_url)
        service = sso_service.registrations.request_password_reset_token
        try:
            result = service(email=email)
        except HTTPError, e:
            logger.exception('request_password_reset_token failed with:')
            raise ResetPasswordTokenError(e.content)

        if result['status'] == 'ok':
            return True
        else:
            raise ResetPasswordTokenError('Received invalid reply: %s' %
                                          result)

    def set_new_password(self, email, token, new_password):
        """Set a new password for the account 'email' to be 'new_password'.

        The 'token' has to be the one resulting from a call to
        'request_password_reset_token'.

        """
        sso_service = self.sso_service_class(None, self.service_url)
        service = sso_service.registrations.set_new_password
        try:
            result = service(email=email, token=token,
                             new_password=new_password)
        except HTTPError, e:
            logger.exception('set_new_password failed with:')
            raise NewPasswordError(e.content)

        if result['status'] == 'ok':
            return True
        else:
            raise NewPasswordError('Received invalid reply: %s' % result)


def blocking(f, result_cb, error_cb):
    """Run f in a thread; return or throw an exception thru the callbacks."""
    def _in_thread():
        """The part that runs inside the thread."""
        try:
            result_cb(f())
        except Exception, e:
            msg = "Exception while running DBus blocking code in a thread."""
            logger.exception(msg)
            error_cb(str(e))
    threading.Thread(target=_in_thread).start()


class SSOLogin(dbus.service.Object):
    """Login thru the Single Sign On service."""

    def __init__(self, bus_name,
                 sso_login_processor_class=SSOLoginProcessor,
                 sso_service_class=None):
        """Initiate the Login object."""
        dbus.service.Object.__init__(self, object_path="/sso",
                                     bus_name=bus_name)
        self.sso_login_processor_class = sso_login_processor_class
        self.sso_service_class = sso_service_class

    def processor(self):
        """Create a login processor with the given class and service class."""
        return self.sso_login_processor_class(
                                    sso_service_class=self.sso_service_class)

    # generate_capcha signals
    @dbus.service.signal(DBUS_IFACE_USER_NAME, signature="s")
    def CaptchaGenerated(self, result):
        """Signal thrown after the captcha is generated."""

    @dbus.service.signal(DBUS_IFACE_USER_NAME, signature="s")
    def CaptchaGenerationError(self, error):
        """Signal thrown when there's a problem generating the captcha."""

    @dbus.service.method(dbus_interface=DBUS_IFACE_USER_NAME,
                         in_signature='s')
    def generate_captcha(self, filename):
        """Call the matching method in the processor."""
        def f():
            """Inner function that will be run in a thread."""
            return self.processor().generate_captcha(filename)
        blocking(f, self.CaptchaGenerated, self.CaptchaGenerationError)

    # register_user signals
    @dbus.service.signal(DBUS_IFACE_USER_NAME, signature="b")
    def UserRegistered(self, result):
        """Signal thrown when the user is registered."""

    @dbus.service.signal(DBUS_IFACE_USER_NAME, signature="s")
    def UserRegistrationError(self, error):
        """Signal thrown when there's a problem registering the user."""

    @dbus.service.method(dbus_interface=DBUS_IFACE_USER_NAME,
                         in_signature='ssss')
    def register_user(self, email, password, captcha_id, captcha_solution):
        """Call the matching method in the processor."""
        def f():
            """Inner function that will be run in a thread."""
            return self.processor().register_user(email, password,
                                              captcha_id, captcha_solution)
        blocking(f, self.UserRegistered, self.UserRegistrationError)

    # login signals
    @dbus.service.signal(DBUS_IFACE_USER_NAME, signature="s")
    def LoggedIn(self, result):
        """Signal thrown when the user is logged in."""

    @dbus.service.signal(DBUS_IFACE_USER_NAME, signature="s")
    def LoginError(self, error):
        """Signal thrown when there is a problem in the login."""

    @dbus.service.method(dbus_interface=DBUS_IFACE_USER_NAME,
                         in_signature='sss')
    def login(self, email, password, app_name):
        """Call the matching method in the processor."""
        def f():
            """Inner function that will be run in a thread."""
            credentials = self.processor().login(email, password, app_name)
            assert credentials is not None
            return app_name
        blocking(f, self.LoggedIn, self.LoginError)

    # validate_email signals
    @dbus.service.signal(DBUS_IFACE_USER_NAME, signature="s")
    def EmailValidated(self, app_name):
        """Signal thrown after the email is validated."""

    @dbus.service.signal(DBUS_IFACE_USER_NAME, signature="s")
    def EmailValidationError(self, error):
        """Signal thrown when there's a problem validating the email."""

    @dbus.service.method(dbus_interface=DBUS_IFACE_USER_NAME,
                         in_signature='ssss')
    def validate_email(self, email, password, app_name, email_token):
        """Call the matching method in the processor."""
        def f():
            """Inner function that will be run in a thread."""
            self.processor().validate_email(email, password,
                                               app_name, email_token)
            return app_name
        blocking(f, self.EmailValidated, self.EmailValidationError)

    # request_password_reset_token signals
    @dbus.service.signal(DBUS_IFACE_USER_NAME, signature="s")
    def PasswordResetTokenSent(self, email):
        """Signal thrown when the token is succesfully sent."""

    @dbus.service.signal(DBUS_IFACE_USER_NAME, signature="s")
    def PasswordResetError(self, error):
        """Signal thrown when there's a problem sending the token."""

    @dbus.service.method(dbus_interface=DBUS_IFACE_USER_NAME,
                         in_signature='s')
    def request_password_reset_token(self, email):
        """Call the matching method in the processor."""
        def f():
            """Inner function that will be run in a thread."""
            self.processor().request_password_reset_token(email)
            return email
        blocking(f, self.PasswordResetTokenSent, self.PasswordResetError)

    # set_new_password signals
    @dbus.service.signal(DBUS_IFACE_USER_NAME, signature="s")
    def PasswordChanged(self, email):
        """Signal thrown when the token is succesfully sent."""

    @dbus.service.signal(DBUS_IFACE_USER_NAME, signature="s")
    def PasswordChangeError(self, error):
        """Signal thrown when there's a problem sending the token."""

    @dbus.service.method(dbus_interface=DBUS_IFACE_USER_NAME,
                         in_signature='sss')
    def set_new_password(self, email, token, new_password):
        """Call the matching method in the processor."""
        def f():
            """Inner function that will be run in a thread."""
            self.processor().set_new_password(email, token, new_password)
            return email
        blocking(f, self.PasswordChanged, self.PasswordChangeError)


class SSOCredentials(dbus.service.Object):
    """DBus object that gets credentials, and login/registers if needed."""

    @dbus.service.signal(DBUS_IFACE_CRED_NAME, signature="")
    def AuthorizationDenied(self):
        """Signal thrown when the user denies the authorization."""

    @dbus.service.signal(DBUS_IFACE_CRED_NAME, signature="a{ss}")
    def CredentialsFound(self, app_name):
        """Signal thrown when the credentials are found."""

    @dbus.service.signal(DBUS_IFACE_CRED_NAME, signature="s")
    def CredentialsError(self, error):
        """Signal thrown when there is a problem finding the credentials."""

    @dbus.service.method(dbus_interface=DBUS_IFACE_CRED_NAME,
                         in_signature="s", out_signature="a{ss}")
    def find_credentials(self, app_name):
        """Get the credentials from the keyring or '' if not there."""
        token = keyring_get_credentials(app_name)
        if token is None:
            return {}
        else:
            return token

    def _login_success_cb(self, dialog, app_name):
        """Handles the response from the UI dialog."""
        try:
            creds = keyring_get_credentials(app_name)
            self.CredentialsFound(creds)
        except Exception, e:
            logger.exception("problem when getting credentials from keyring")
            self.CredentialsError(str(e))

    def _login_error_cb(self, dialog):
        """Handles a problem in the UI."""
        msg = "misc problem getting the credentials"
        self.CredentialsError(msg)
        logger.error(msg)

    def _login_auth_denied_cb(self, dialog):
        """The user decides not to allow the registration or login."""
        self.AuthorizationDenied()

    def _show_login_or_register_ui(self, app_name, tc_url, help_text,
                                   win_id, login_only=False):
        """Shows the UI so the user can login or register."""
        try:
            gui_app = gui.UbuntuSSOClientGUI(app_name, tc_url,
                                             help_text, win_id, login_only)
            gui_app.connect(gui.SIG_LOGIN_SUCCEEDED, self._login_success_cb)
            gui_app.connect(gui.SIG_LOGIN_FAILED, self._login_error_cb)
            gui_app.connect(gui.SIG_REGISTRATION_SUCCEEDED,
                            self._login_success_cb)
            gui_app.connect(gui.SIG_REGISTRATION_FAILED, self._login_error_cb)
            gui_app.connect(gui.SIG_USER_CANCELATION,
                            self._login_auth_denied_cb)
        except Exception, e:
            msg = '_show_login_or_register_ui failed when calling ' \
                  'gui.UbuntuSSOClientGUI(%r, %r, %r, %r, %r)'
            logger.exception(msg, app_name, tc_url, help_text,
                             win_id, login_only)
            self.CredentialsError(str(e))

    @dbus.service.method(dbus_interface=DBUS_IFACE_CRED_NAME,
                         in_signature="sssx", out_signature="")
    def login_or_register_to_get_credentials(self, app_name,
                                             terms_and_conditions_url,
                                             help_text, window_id):
        """Get credentials if found else prompt GUI to login or register.

        'app_name' will be displayed in the GUI.
        'terms_and_conditions_url' will be the URL pointing to T&C.
        'help_text' is an explanatory text for the end-users, will be shown
         below the headers.
        'window_id' is the id of the window which will be set as a parent of
         the GUI. If 0, no parent will be set.

        """
        try:
            token = keyring_get_credentials(app_name)
            if token is None:
                gobject.idle_add(self._show_login_or_register_ui,
                                 app_name, terms_and_conditions_url,
                                 help_text, window_id)
            else:
                self.CredentialsFound(token)
        except Exception, e:
            logger.exception("problem while getting the credentials")
            self.CredentialsError(str(e))

    def _show_login_only_ui(self, app_name, help_text, win_id):
        """Shows the UI so the user can login."""
        tc_url = None  # no terms and conditions when only logging in
        self._show_login_or_register_ui(app_name, tc_url, help_text,
                                        win_id, login_only=True)

    @dbus.service.method(dbus_interface=DBUS_IFACE_CRED_NAME,
                         in_signature="ssx", out_signature="")
    def login_to_get_credentials(self, app_name, help_text, window_id):
        """Get credentials if found else prompt GUI just to login

        'app_name' will be displayed in the GUI.
        'help_text' is an explanatory text for the end-users, will be shown
         before the login fields.
        'window_id' is the id of the window which will be set as a parent of
         the GUI. If 0, no parent will be set.

        """
        try:
            token = keyring_get_credentials(app_name)
            if token is None:
                gobject.idle_add(self._show_login_only_ui, app_name,
                                 help_text, window_id)
            else:
                self.CredentialsFound(token)
        except Exception, e:
            logger.exception("problem while getting the credentials")
            self.CredentialsError(str(e))


class LoginProcessor:
    """Actually do the work of processing passed parameters."""

    def __init__(self, dbus_object):
        """Initialize the login processor."""
        logger.debug("Creating a LoginProcessor")
        self.note1 = None
        self.realm = None
        self.consumer_key = None
        self.dbus_object = dbus_object
        logger.debug("Getting configuration")
        self.config = get_config()

    def login(self, realm, consumer_key, do_login=True):
        """Initiate an OAuth login"""
        logger.debug("Initiating OAuth login in LoginProcessor")
        self.realm = str(realm)  # because they are dbus.Strings, not str
        self.consumer_key = str(consumer_key)

        logger.debug("Obtaining OAuth urls")
        (request_token_url, user_authorisation_url,
          access_token_url, consumer_secret) = self.get_config_urls(realm)
        logger.debug("OAuth URLs are: request='%s', userauth='%s', " + \
                     "access='%s', secret='%s'", request_token_url,
                     user_authorisation_url, access_token_url, consumer_secret)

        from ubuntu_sso.auth import AuthorisationClient
        client = AuthorisationClient(self.realm,
                                     request_token_url,
                                     user_authorisation_url,
                                     access_token_url, self.consumer_key,
                                     consumer_secret,
                                     callback_parent=self.got_token,
                                     callback_denied=self.got_denial,
                                     callback_notoken=self.got_no_token,
                                     callback_error=self.got_error,
                                     do_login=do_login)

        logger.debug("Calling auth.client.ensure_access_token in thread")
        gobject.timeout_add_seconds(1, client.ensure_access_token)

    def clear_token(self, realm, consumer_key):
        """Remove the currently stored OAuth token from the keyring."""
        self.realm = str(realm)
        self.consumer_key = str(consumer_key)
        (request_token_url, user_authorisation_url,
          access_token_url, consumer_secret) = self.get_config_urls(self.realm)
        from ubuntu_sso.auth import AuthorisationClient
        client = AuthorisationClient(self.realm,
                                     request_token_url,
                                     user_authorisation_url,
                                     access_token_url,
                                     self.consumer_key, consumer_secret,
                                     callback_parent=self.got_token,
                                     callback_denied=self.got_denial,
                                     callback_notoken=self.got_no_token,
                                     callback_error=self.got_error)
        gobject.timeout_add_seconds(1, client.clear_token)

    def error_handler(self, failure):
        """Deal with errors returned from auth process"""
        logger.debug("Error returned from auth process")
        self.dbus_object.currently_authing = False  # not block future requests

    def get_config_urls(self, realm):
        """Look up the URLs to use in the config file"""
        logger.debug("Fetching config URLs for realm='%s'", realm)
        if self.config.has_section(realm):
            logger.debug("Realm '%s' is in config", realm)
            request_token_url = self.__get_url(realm, "request_token_url")
            user_authorisation_url = self.__get_url(realm,
              "user_authorisation_url")
            access_token_url = self.__get_url(realm, "access_token_url")
            consumer_secret = self.__get_option(realm, "consumer_secret")
        elif realm.startswith("http://localhost") and \
          self.config.has_section("http://localhost"):
            logger.debug("Realm is localhost and is in config")
            request_token_url = self.__get_url("http://localhost",
              "request_token_url", realm)
            user_authorisation_url = self.__get_url("http://localhost",
              "user_authorisation_url", realm)
            access_token_url = self.__get_url("http://localhost",
              "access_token_url", realm)
            consumer_secret = self.__get_option("http://localhost",
              "consumer_secret")
        elif self.is_valid_url(realm):
            logger.debug("Realm '%s' is not in config", realm)
            request_token_url = self.__get_url("default",
              "request_token_url", realm)
            user_authorisation_url = self.__get_url("default",
              "user_authorisation_url", realm)
            access_token_url = self.__get_url("default",
              "access_token_url", realm)
            consumer_secret = self.__get_option(realm, "consumer_secret")
        else:
            logger.debug("Realm '%s' is a bad realm", realm)
            raise BadRealmError
        return (request_token_url, user_authorisation_url,
                access_token_url, consumer_secret)

    def is_valid_url(self, url):
        """Simple check for URL validity"""
        # pylint: disable-msg=W0612
        scheme, netloc, path, query, fragment = urlparse.urlsplit(url)
        if scheme and netloc:
            return True
        else:
            return False

    def got_token(self, access_token):
        """Callback function when access token has been retrieved"""
        logger.debug("Token retrieved, calling NewCredentials function")
        self.dbus_object.NewCredentials(self.realm, self.consumer_key)

    def got_denial(self):
        """Callback function when request token has been denied"""
        self.dbus_object.AuthorizationDenied()

    def got_no_token(self):
        """Callback function when access token is not in keyring."""
        self.dbus_object.NoCredentials()

    def got_error(self, message):
        """Callback function to emit an error message over DBus."""
        self.dbus_object.OAuthError(message)

    def __get_url(self, realm, option, actual_realm=None):
        """Construct a full URL from realm and a URLpath for that realm in
           the config file."""
        if actual_realm:
            realm_to_use = actual_realm
        else:
            realm_to_use = realm
        urlstub = self.__get_option(realm, option)
        return urlparse.urljoin(realm_to_use, urlstub)

    def __get_option(self, realm, option):
        """Return a specific option for that realm in
           the config file. If the realm does not exist in the config file,
           fall back to the [default] section."""
        if self.config.has_section(realm) and \
           self.config.has_option(realm, option):
            urlstub = self.config.get(realm, option)
            return urlstub

        # either the realm exists and this url does not, or
        # the realm doesn't exist; either way, fall back to [default] section
        urlstub = self.config.get("default", option, None)
        if urlstub is not None:
            return urlstub

        # this url does not exist in default section either
        # this shouldn't happen
        raise NoDefaultConfigError("No default configuration for %s" % option)


class Login(dbus.service.Object):
    """Object which listens for D-Bus OAuth requests."""

    def __init__(self, bus_name):
        """Initiate the Login object."""
        dbus.service.Object.__init__(self, object_path="/", bus_name=bus_name)
        self.processor = LoginProcessor(self)
        self.currently_authing = False
        logger.debug("Login D-Bus service starting up")

    @dbus.service.method(dbus_interface=DBUS_IFACE_AUTH_NAME,
                         in_signature='ss', out_signature='')
    def login(self, realm, consumer_key):
        """D-Bus method, to initiate an OAuth login."""
        logger.debug("login() D-Bus message received with realm='%s', " +
                     "consumer_key='%s'", realm, consumer_key)
        if self.currently_authing:
            logger.debug("Currently in the middle of OAuth: rejecting this")
            return
        self.currently_authing = True
        self.processor.login(realm, consumer_key)

    @dbus.service.method(dbus_interface=DBUS_IFACE_AUTH_NAME,
                         in_signature='ssb', out_signature='')
    def maybe_login(self, realm, consumer_key, do_login):
        """D-Bus method, to maybe initiate an OAuth login."""
        logger.debug("maybe_login() D-Bus message received with realm='%s', " +
                     "consumer_key='%s'", realm, consumer_key)
        if self.currently_authing:
            logger.debug("Currently in the middle of OAuth: rejecting this")
            return
        self.currently_authing = True
        self.processor.login(realm, consumer_key, do_login)

    @dbus.service.method(dbus_interface=DBUS_IFACE_AUTH_NAME,
                         in_signature='ss', out_signature='')
    def clear_token(self, realm, consumer_key):
        """D-Bus method, to clear the existing token."""
        self.processor.clear_token(realm, consumer_key)

    @dbus.service.signal(dbus_interface=DBUS_IFACE_AUTH_NAME, signature='ss')
    def NewCredentials(self, realm, consumer_key):
        """Fire D-Bus signal when the user accepts authorization."""
        logger.debug("Firing the NewCredentials signal")
        self.currently_authing = False
        return (self.processor.realm, self.processor.consumer_key)

    @dbus.service.signal(dbus_interface=DBUS_IFACE_AUTH_NAME)
    def AuthorizationDenied(self):
        """Fire the signal when the user denies authorization."""
        self.currently_authing = False

    @dbus.service.signal(dbus_interface=DBUS_IFACE_AUTH_NAME)
    def NoCredentials(self):
        """Fired when the user does not have a token in the keyring."""
        self.currently_authing = False

    @dbus.service.signal(dbus_interface=DBUS_IFACE_AUTH_NAME, signature='s')
    def OAuthError(self, message):
        """Fire the signal when an error needs to be propagated to the user."""
        self.currently_authing = False
        return message


def main():
    """Start everything"""
    dbus.mainloop.glib.threads_init()
    gobject.threads_init()
    logger.debug("Starting up at %s", time.asctime())
    logger.debug("Installing the Twisted glib2reactor")
    from twisted.internet import glib2reactor  # for non-GUI apps
    glib2reactor.install()
    from twisted.internet import reactor

    logger.debug("Creating the D-Bus service")
    Login(dbus.service.BusName(DBUS_BUS_NAME,
                               bus=dbus.SessionBus()))
    SSOLogin(dbus.service.BusName(DBUS_BUS_NAME,
                               bus=dbus.SessionBus()))
    SSOCredentials(dbus.service.BusName(DBUS_BUS_NAME,
                               bus=dbus.SessionBus()),
                   object_path=DBUS_CRED_PATH)
    # cleverness here to say:
    # am I already running (bound to this d-bus name)?
    # if so, send a signal to the already running instance
    # this means that this app can be started from an x-ubutnuone: URL
    # to kick off the signin process
    logger.debug("Starting the reactor mainloop")
    reactor.run()
