# test_keyring - tests for ubuntu_sso.keyring
#
# Author: Alejandro J. Cura <alecu@canonical.com>
#
# Copyright 2010 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Tests for the keyring.py module."""

import gnomekeyring
from twisted.trial.unittest import TestCase

from ubuntu_sso import keyring


class MockInfoSync(object):
    """A mock object that represents a fake key."""

    def __init__(self, (key, key_name, secret)):
        """Initialize this instance."""
        self.key = key
        self.key_name = key_name
        self.secret = secret

    def get_display_name(self):
        """Return info on the mocked object."""
        return self.key_name


class MockGnomeKeyring(object):
    """A mock keyring that stores keys according to a given attr."""

    def __init__(self):
        """Initialize this instance."""
        self.keyrings = set()
        self.store = []
        self.deleted = []

    def item_create_sync(self, keyring_name, item_type, key_name, attr,
                         secret, update_if_exists):
        """Sets a value in the keyring."""
        # we'll use the attr as the dict key, so make it a tuple
        key = tuple(attr.items())
        self.store.append((key, key_name, secret))

    def item_delete_sync(self, keyring_name, item_id):
        """Makes a list of deleted items."""
        key, key_name, secret = self.store.pop(item_id)
        self.deleted.append(key_name)

    def list_item_ids_sync(self, keyring_name):
        """Return a list of ids for all the items."""
        return [n for n, v in enumerate(self.store)]

    def item_get_info_sync(self, keyring_name, item_id):
        """Return an info sync object for the item."""
        return MockInfoSync(self.store[item_id])

    def list_keyring_names_sync(self):
        """The keyring you are looking for may be here."""
        return self.keyrings

    def create_sync(self, name):
        """Thanks, I'll create that keyring for you."""
        self.keyrings.add(name)

    def is_available(self):
        """A very available keyring."""
        return True


class TestKeyring(TestCase):
    """Test the gnome keyring related functions."""
    def setUp(self):
        """Initialize the mock used in these tests."""
        self.mgk = MockGnomeKeyring()
        self.patch(gnomekeyring, "item_create_sync", self.mgk.item_create_sync)
        self.patch(gnomekeyring, "is_available", self.mgk.is_available)
        self.patch(gnomekeyring, "create_sync", self.mgk.create_sync)
        self.patch(gnomekeyring, "list_keyring_names_sync",
                                              self.mgk.list_keyring_names_sync)
        self.patch(gnomekeyring, "list_item_ids_sync",
                                                   self.mgk.list_item_ids_sync)
        self.patch(gnomekeyring, "item_get_info_sync",
                                                   self.mgk.item_get_info_sync)
        self.patch(gnomekeyring, "item_delete_sync", self.mgk.item_delete_sync)

    def test_set_ubuntusso(self):
        """Test that the set method does not erase previous keys."""
        sample_creds = {"name": "sample creds name"}
        sample_creds2 = {"name": "sample creds name 2"}
        keyring.Keyring("appname").set_ubuntusso_attr(sample_creds)
        keyring.Keyring("appname").set_ubuntusso_attr(sample_creds2)

        self.assertEqual(len(self.mgk.store), 2)
        self.assertEqual(len(self.mgk.deleted), 0)

    def test_delete_ubuntusso(self):
        """Test that a given key is deleted."""
        sample_creds = {"name": "sample creds name"}
        keyring.Keyring("appname").set_ubuntusso_attr(sample_creds)
        keyring.Keyring("appname").delete_ubuntusso_attr()

        self.assertEqual(len(self.mgk.store), 0)
        self.assertEqual(len(self.mgk.deleted), 1)
