# test_main - tests for ubuntu_sso.main
#
# Author: Stuart Langridge <stuart.langridge@canonical.com>
# Author: Natalia Bidart <natalia.bidart@canonical.com>
# Author: Alejandro J. Cura <alecu@canonical.com>
#
# Copyright 2009-2010 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Tests for the OAuth client code."""

import logging
import os
import StringIO

import gobject

from lazr.restfulclient.errors import HTTPError
from mocker import Mocker, MockerTestCase, ARGS, KWARGS
from twisted.internet.defer import Deferred
from twisted.trial.unittest import TestCase

import ubuntu_sso.main

from contrib.testing.testcase import MementoHandler
from ubuntu_sso import config, gui
from ubuntu_sso.main import (
    AuthenticationError, BadRealmError, blocking, EmailTokenError,
    InvalidEmailError, InvalidPasswordError, SSOLogin, SSOCredentials,
    get_token_name, keyring_get_credentials, keyring_store_credentials, logger,
    LoginProcessor, NewPasswordError, SSOLoginProcessor, RegistrationError,
    ResetPasswordTokenError, OLD_KEY_NAME, U1_APP_NAME)


APP_NAME = 'The Coolest App Ever'
CAPTCHA_PATH = os.path.abspath(os.path.join(os.curdir, 'ubuntu_sso', 'tests',
                                            'files', 'captcha.png'))
CAPTCHA_ID = 'test'
CAPTCHA_SOLUTION = 'william Byrd'
CANT_RESET_PASSWORD_CONTENT = "CanNotResetPassowrdError: " \
                              "Can't reset password for this account"
RESET_TOKEN_INVALID_CONTENT = "AuthToken matching query does not exist."
EMAIL = 'test@example.com'
EMAIL_TOKEN = 'B2Pgtf'
HELP = 'help text'
PASSWORD = 'be4tiFul'
RESET_PASSWORD_TOKEN = '8G5Wtq'
TOKEN = {u'consumer_key': u'xQ7xDAz',
         u'consumer_secret': u'KzCJWCTNbbntwfyCKKjomJDzlgqxLy',
         u'name': u'test',
         u'token': u'GkInOfSMGwTXAUoVQwLUoPxElEEUdhsLVNTPhxHJDUIeHCPNEo',
         u'token_secret': u'qFYImEtlczPbsCnYyuwLoPDlPEnvNcIktZphPQklAWrvyfFMV'}
STATUS_UNKNOWN = {'status': 'yadda-yadda'}
STATUS_ERROR = {'status': 'error', 'errors': {'something': ['Bla', 'Ble']}}
STATUS_OK = {'status': 'ok'}
STATUS_EMAIL_UNKNOWN = {'status': 'yadda-yadda'}
STATUS_EMAIL_ERROR = {'errors': {'email_token': ['Error1', 'Error2']}}
STATUS_EMAIL_OK = {'email': 'okmail@okserver.okdomain'}
TC_URL = 'tcurl'
WINDOW_ID = 5

LOGIN_OR_REGISTER_ARGS = (APP_NAME, TC_URL, HELP, WINDOW_ID)
LOGIN_OR_REGISTER_GUI_ARGS = LOGIN_OR_REGISTER_ARGS + (False,)
LOGIN_ONLY_ARGS = (APP_NAME, HELP, WINDOW_ID)
LOGIN_ONLY_GUI_ARGS = (APP_NAME, None, HELP, WINDOW_ID, True)


class FakedCaptchas(object):
    """Fake the captcha generator."""

    def new(self):
        """Return a fix captcha)."""
        return {'image_url': 'file://%s' % CAPTCHA_PATH,
                'captcha_id': CAPTCHA_ID}


class FakedRegistrations(object):
    """Fake the registrations service."""

    def register(self, email, password, captcha_id, captcha_solution):
        """Fake registration. Return a fix result."""
        if captcha_id is None and captcha_solution is None:
            return STATUS_UNKNOWN
        elif captcha_id != CAPTCHA_ID or captcha_solution != CAPTCHA_SOLUTION:
            return STATUS_ERROR
        else:
            return STATUS_OK

    def request_password_reset_token(self, email):
        """Fake password reset token. Return a fix result."""
        if email is None:
            return STATUS_UNKNOWN
        elif email != EMAIL:
            raise HTTPError(response=None, content=CANT_RESET_PASSWORD_CONTENT)
        else:
            return STATUS_OK

    def set_new_password(self, email, token, new_password):
        """Fake the setting of new password. Return a fix result."""
        if email is None and token is None and new_password is None:
            return STATUS_UNKNOWN
        elif email != EMAIL or token != RESET_PASSWORD_TOKEN:
            raise HTTPError(response=None, content=RESET_TOKEN_INVALID_CONTENT)
        else:
            return STATUS_OK


class FakedAuthentications(object):
    """Fake the authentications service."""

    def authenticate(self, token_name):
        """Fake authenticate. Return a fix result."""
        if not token_name.startswith(get_token_name(APP_NAME)):
            raise HTTPError(response=None, content=None)
        else:
            return TOKEN


class FakedAccounts(object):
    """Fake the accounts service."""

    def validate_email(self, email_token):
        """Fake the email validation. Return a fix result."""
        if email_token is None:
            return STATUS_EMAIL_UNKNOWN
        if email_token != EMAIL_TOKEN:
            return STATUS_EMAIL_ERROR
        else:
            return STATUS_EMAIL_OK


class FakedSSOServer(object):
    """Fake an SSO server."""

    def __init__(self, authorizer, service_root):
        self.captchas = FakedCaptchas()
        self.registrations = FakedRegistrations()
        self.authentications = FakedAuthentications()
        self.accounts = FakedAccounts()


class SSOLoginProcessorTestCase(TestCase, MockerTestCase):
    """Test suite for the SSO login processor."""

    def setUp(self):
        """Init."""
        self.processor = SSOLoginProcessor(sso_service_class=FakedSSOServer)
        self.register_kwargs = dict(email=EMAIL, password=PASSWORD,
                                    captcha_id=CAPTCHA_ID,
                                    captcha_solution=CAPTCHA_SOLUTION)
        self.login_kwargs = dict(email=EMAIL, password=PASSWORD,
                                 app_name=APP_NAME)

        def ksc(k, v):
            self.assertEqual(k, APP_NAME)
            self.assertEqual(v, TOKEN)

        self.patch(ubuntu_sso.main, "keyring_store_credentials", ksc)

    def tearDown(self):
        """Clean up."""
        self.processor = None

    def test_generate_captcha(self):
        """Captcha can be generated."""
        filename = self.mktemp()
        self.addCleanup(lambda: os.remove(filename))
        captcha_id = self.processor.generate_captcha(filename)
        self.assertEqual(CAPTCHA_ID, captcha_id, 'captcha id must be correct.')
        self.assertTrue(os.path.isfile(filename), '%s must exist.' % filename)

        with open(CAPTCHA_PATH) as f:
            expected = f.read()
        with open(filename) as f:
            actual = f.read()
        self.assertEqual(expected, actual, 'captcha image must be correct.')

    def test_register_user_checks_valid_email(self):
        """Email is validated."""
        self.register_kwargs['email'] = 'notavalidemail'
        self.assertRaises(InvalidEmailError,
                          self.processor.register_user, **self.register_kwargs)

    def test_register_user_checks_valid_password(self):
        """Password is validated."""
        self.register_kwargs['password'] = ''
        self.assertRaises(InvalidPasswordError,
                          self.processor.register_user, **self.register_kwargs)

        # 7 chars, one less than expected
        self.register_kwargs['password'] = 'tesT3it'
        self.assertRaises(InvalidPasswordError,
                          self.processor.register_user, **self.register_kwargs)

        self.register_kwargs['password'] = 'test3it!'  # no upper case
        self.assertRaises(InvalidPasswordError,
                          self.processor.register_user, **self.register_kwargs)

        self.register_kwargs['password'] = 'testIt!!'  # no number
        self.assertRaises(InvalidPasswordError,
                          self.processor.register_user, **self.register_kwargs)

    # register

    def test_register_user_if_status_ok(self):
        """A user is succesfuy registered into the SSO server."""
        result = self.processor.register_user(**self.register_kwargs)
        self.assertTrue(result, 'registration was successful.')

    def test_register_user_if_status_error(self):
        """Proper error is raised if register fails."""
        self.register_kwargs['captcha_id'] = CAPTCHA_ID * 2  # incorrect
        failure = self.assertRaises(RegistrationError,
                                    self.processor.register_user,
                                    **self.register_kwargs)
        self.assertIn(STATUS_ERROR['errors'], failure)

    def test_register_user_if_status_unknown(self):
        """Proper error is raised if register returns an unknown status."""
        self.register_kwargs['captcha_id'] = None
        self.register_kwargs['captcha_solution'] = None
        failure = self.assertRaises(RegistrationError,
                                    self.processor.register_user,
                                    **self.register_kwargs)
        self.assertIn('Received unknown status: %s' % STATUS_UNKNOWN, failure)

    # login

    def test_login_if_http_error(self):
        """Proper error is raised if authentication fails."""
        self.login_kwargs['app_name'] = APP_NAME * 2  # invalid token name
        self.assertRaises(AuthenticationError,
                          self.processor.login, **self.login_kwargs)

    def test_login_if_no_error(self):
        """A user can be succesfully logged in into the SSO service."""
        result = self.processor.login(**self.login_kwargs)
        self.assertEqual(TOKEN, result, 'authentication was successful.')

    # validate_email

    def test_validate_email_if_status_ok(self):
        """A email is succesfuy validated in the SSO server."""
        self.login_kwargs['email_token'] = EMAIL_TOKEN  # valid email token
        result = self.processor.validate_email(**self.login_kwargs)
        self.assertTrue(result, 'email validation was successful.')

    def test_validate_email_if_status_error(self):
        """Proper error is raised if email validation fails."""
        self.login_kwargs['email_token'] = EMAIL_TOKEN * 2  # invalid token
        failure = self.assertRaises(EmailTokenError,
                                    self.processor.validate_email,
                                    **self.login_kwargs)
        self.assertIn(STATUS_EMAIL_ERROR['errors'], failure)

    def test_validate_email_if_status_unknown(self):
        """Proper error is raised if email validation returns unknown."""
        self.login_kwargs['email_token'] = None
        failure = self.assertRaises(EmailTokenError,
                                    self.processor.validate_email,
                                    **self.login_kwargs)
        self.assertIn('Received invalid reply: %s' % STATUS_UNKNOWN, failure)

    # reset_password

    def test_request_password_reset_token_if_status_ok(self):
        """A reset password token is succesfuly sent."""
        result = self.processor.request_password_reset_token(email=EMAIL)
        self.assertTrue(result, 'password reset token must be successful.')

    def test_request_password_reset_token_if_http_error(self):
        """Proper error is raised if password token request fails."""
        exc = self.assertRaises(ResetPasswordTokenError,
                                self.processor.request_password_reset_token,
                                email=EMAIL * 2)
        self.assertIn(CANT_RESET_PASSWORD_CONTENT, exc)

    def test_request_password_reset_token_if_status_unknown(self):
        """Proper error is raised if password token request returns unknown."""
        exc = self.assertRaises(ResetPasswordTokenError,
                                self.processor.request_password_reset_token,
                                email=None)
        self.assertIn('Received invalid reply: %s' % STATUS_UNKNOWN, exc)

    def test_set_new_password_if_status_ok(self):
        """A new password is succesfuy set."""
        result = self.processor.set_new_password(email=EMAIL,
                                                 token=RESET_PASSWORD_TOKEN,
                                                 new_password=PASSWORD)
        self.assertTrue(result, 'new password must be set successfully.')

    def test_set_new_password_if_http_error(self):
        """Proper error is raised if setting a new password fails."""
        exc = self.assertRaises(NewPasswordError,
                                self.processor.set_new_password,
                                email=EMAIL * 2,
                                token=RESET_PASSWORD_TOKEN * 2,
                                new_password=PASSWORD)
        self.assertIn(RESET_TOKEN_INVALID_CONTENT, exc)

    def test_set_new_password_if_status_unknown(self):
        """Proper error is raised if setting a new password returns unknown."""
        exc = self.assertRaises(NewPasswordError,
                                self.processor.set_new_password,
                                email=None, token=None, new_password=None)
        self.assertIn('Received invalid reply: %s' % STATUS_UNKNOWN, exc)


class Realm(MockerTestCase):
    """Test the realm handling finder"""

    def test_invalid_realm(self):
        """Are invalid realms rejected?"""
        login = LoginProcessor(None)
        self.assertRaises(BadRealmError, login.login, "bad realm", "key")

    def test_realms(self):
        """Are realm URLs correctly retrieved from the config?"""
        # mock that tmp config file exists
        tmp = os.path.realpath(os.path.join(os.path.split(config.__file__)[0],
                                            "../../data/oauth_urls"))
        osp = self.mocker.replace("os.path")
        osp.isfile(tmp)
        self.mocker.result(True)

        sio = StringIO.StringIO("""[default]
request_token_url = /rtu-default
user_authorisation_url = /uau-default
access_token_url = /atu-default
consumer_secret = foo-default

[http://localhost]
request_token_url = /rtu-localhost
user_authorisation_url = /uau-localhost
access_token_url = /atu-localhost
consumer_secret = foo-localhost

[http://ubuntuone.com]
request_token_url = /rtu-ubuntuone
user_authorisation_url = /uau-ubuntuone
access_token_url = /atu-ubuntuone
consumer_secret = foo-ubuntuone

""")
        mock_open = self.mocker.replace(open)
        mock_open(tmp)
        self.mocker.result(sio)

        self.mocker.replay()

        login = LoginProcessor(None)

        # are localhost:XXXX URLs correctly fetched?
        (rtu, uau, atu, cs) = login.get_config_urls("http://localhost:9876")
        self.assertEqual(rtu, "http://localhost:9876/rtu-localhost")
        self.assertEqual(uau, "http://localhost:9876/uau-localhost")
        self.assertEqual(atu, "http://localhost:9876/atu-localhost")
        self.assertEqual(cs, "foo-localhost")

        # is a URL specifically in the config correctly fetched?
        (rtu, uau, atu, cs) = login.get_config_urls("http://ubuntuone.com")
        self.assertEqual(rtu, "http://ubuntuone.com/rtu-ubuntuone")
        self.assertEqual(uau, "http://ubuntuone.com/uau-ubuntuone")
        self.assertEqual(atu, "http://ubuntuone.com/atu-ubuntuone")
        self.assertEqual(cs, "foo-ubuntuone")

        # is a URL not in the config correctly fetched as default?
        (rtu, uau, atu, cs) = login.get_config_urls("http://other.example.net")
        self.assertEqual(rtu, "http://other.example.net/rtu-default")
        self.assertEqual(uau, "http://other.example.net/uau-default")
        self.assertEqual(atu, "http://other.example.net/atu-default")
        self.assertEqual(cs, "foo-default")


class BlockingSampleException(Exception):
    """The exception that will be thrown by the fake blocking."""


class SsoDbusTestCase(TestCase):
    """Test the SSOLogin DBus interface."""

    def setUp(self):
        """Create the mocking bus."""
        self.real_blocking = ubuntu_sso.main.blocking
        self.mocker = Mocker()
        self.mockbusname = self.mocker.mock()
        mockbus = self.mocker.mock()
        self.mockbusname.get_bus()
        self.mocker.result(mockbus)
        mockbus._register_object_path(ARGS)

    def tearDown(self):
        """Verify the mocking bus and shut it down."""
        self.mocker.verify()
        self.mocker.restore()

    def fake_ok_blocking(self, f, cb, eb):
        """A fake blocking function that succeeds."""
        cb(f())

    def fake_err_blocking(self, f, cb, eb):
        """A fake blocking function that fails."""
        f()
        eb(BlockingSampleException())

    def test_creation(self):
        """Test that the object creation is successful."""
        self.mocker.replay()
        SSOLogin(self.mockbusname)

    def create_mock_processor(self):
        """Create a mock processor from a dummy processor class."""
        self.mockprocessorclass = self.mocker.mock()
        mockprocessor = self.mocker.mock()
        self.mockprocessorclass(ARGS, KWARGS)
        self.mocker.result(mockprocessor)
        return mockprocessor

    def test_generate_captcha(self):
        """Test that the captcha method works ok."""
        d = Deferred()
        filename = "sample filename"
        expected_result = "expected result"
        self.create_mock_processor().generate_captcha(filename)
        self.mocker.result(expected_result)
        self.patch(ubuntu_sso.main, "blocking", self.fake_ok_blocking)
        self.mocker.replay()

        def verify(result):
            self.assertEqual(result, expected_result)
            d.callback(result)

        client = SSOLogin(self.mockbusname,
                          sso_login_processor_class=self.mockprocessorclass)
        self.patch(client, "CaptchaGenerated", verify)
        self.patch(client, "CaptchaGenerationError", d.errback)
        client.generate_captcha(filename)
        return d

    def test_generate_captcha_error(self):
        """Test that the captcha method fails as expected."""
        d = Deferred()
        filename = "sample filename"
        expected_result = "expected result"
        self.create_mock_processor().generate_captcha(filename)
        self.mocker.result(expected_result)
        self.patch(ubuntu_sso.main, "blocking", self.fake_err_blocking)
        self.mocker.replay()

        def verify(error):
            self.assertIsInstance(error, BlockingSampleException)
            d.callback("Ok")

        client = SSOLogin(self.mockbusname,
                          sso_login_processor_class=self.mockprocessorclass)
        self.patch(client, "CaptchaGenerated", d.errback)
        self.patch(client, "CaptchaGenerationError", verify)
        client.generate_captcha(filename)
        return d

    def test_register_user(self):
        """Test that the register_user method works ok."""
        d = Deferred()
        expected_result = "expected result"
        self.create_mock_processor().register_user(EMAIL, PASSWORD, CAPTCHA_ID,
                                                   CAPTCHA_SOLUTION)
        self.mocker.result(expected_result)
        self.patch(ubuntu_sso.main, "blocking", self.fake_ok_blocking)
        self.mocker.replay()

        def verify(result):
            self.assertEqual(result, expected_result)
            d.callback(result)

        client = SSOLogin(self.mockbusname,
                          sso_login_processor_class=self.mockprocessorclass)
        self.patch(client, "UserRegistered", verify)
        self.patch(client, "UserRegistrationError", d.errback)
        client.register_user(EMAIL, PASSWORD, CAPTCHA_ID, CAPTCHA_SOLUTION)
        return d

    def test_register_user_error(self):
        """Test that the register_user method fails as expected."""
        d = Deferred()
        expected_result = "expected result"
        self.create_mock_processor().register_user(EMAIL, PASSWORD, CAPTCHA_ID,
                                                   CAPTCHA_SOLUTION)
        self.mocker.result(expected_result)
        self.patch(ubuntu_sso.main, "blocking", self.fake_err_blocking)
        self.mocker.replay()

        def verify(error):
            self.assertIsInstance(error, BlockingSampleException)
            d.callback("Ok")

        client = SSOLogin(self.mockbusname,
                          sso_login_processor_class=self.mockprocessorclass)
        self.patch(client, "UserRegistered", d.errback)
        self.patch(client, "UserRegistrationError", verify)
        client.register_user(EMAIL, PASSWORD, CAPTCHA_ID, CAPTCHA_SOLUTION)
        return d

    def test_login(self):
        """Test that the login method works ok."""
        d = Deferred()
        sample_result = {"sample": "tokens"}
        self.create_mock_processor().login(EMAIL, PASSWORD, APP_NAME)
        self.mocker.result(sample_result)
        self.patch(ubuntu_sso.main, "blocking", self.fake_ok_blocking)
        self.mocker.replay()

        def verify(result):
            self.assertEqual(result, APP_NAME)
            d.callback(result)

        client = SSOLogin(self.mockbusname,
                          sso_login_processor_class=self.mockprocessorclass)
        self.patch(client, "LoggedIn", verify)
        self.patch(client, "LoginError", d.errback)
        client.login(EMAIL, PASSWORD, APP_NAME)
        return d

    def test_login_error(self):
        """Test that the login method fails as expected."""
        d = Deferred()
        expected_result = "expected result"
        self.create_mock_processor().login(EMAIL, PASSWORD, APP_NAME)
        self.mocker.result(expected_result)
        self.patch(ubuntu_sso.main, "blocking", self.fake_err_blocking)
        self.mocker.replay()

        def verify(error):
            self.assertIsInstance(error, BlockingSampleException)
            d.callback("Ok")

        client = SSOLogin(self.mockbusname,
                          sso_login_processor_class=self.mockprocessorclass)
        self.patch(client, "LoggedIn", d.errback)
        self.patch(client, "LoginError", verify)
        client.login(EMAIL, PASSWORD, APP_NAME)
        return d

    def test_validate_email(self):
        """Test that the validate_email method works ok."""
        d = Deferred()
        self.create_mock_processor().validate_email(EMAIL, PASSWORD,
                                                    APP_NAME, EMAIL_TOKEN)
        self.mocker.result(APP_NAME)
        self.patch(ubuntu_sso.main, "blocking", self.fake_ok_blocking)
        self.mocker.replay()

        def verify(result):
            self.assertEqual(result, APP_NAME)
            d.callback(result)

        client = SSOLogin(self.mockbusname,
                          sso_login_processor_class=self.mockprocessorclass)
        self.patch(client, "EmailValidated", verify)
        self.patch(client, "EmailValidationError", d.errback)
        client.validate_email(EMAIL, PASSWORD, APP_NAME, EMAIL_TOKEN)
        return d

    def test_validate_email_error(self):
        """Test that the validate_email method fails as expected."""
        d = Deferred()
        expected_result = "expected result"

        self.create_mock_processor().validate_email(EMAIL, PASSWORD,
                                                    APP_NAME, EMAIL_TOKEN)
        self.mocker.result(expected_result)
        self.patch(ubuntu_sso.main, "blocking", self.fake_err_blocking)
        self.mocker.replay()

        def verify(error):
            self.assertIsInstance(error, BlockingSampleException)
            d.callback("Ok")

        client = SSOLogin(self.mockbusname,
                          sso_login_processor_class=self.mockprocessorclass)
        self.patch(client, "EmailValidated", d.errback)
        self.patch(client, "EmailValidationError", verify)
        client.validate_email(EMAIL, PASSWORD, APP_NAME, EMAIL_TOKEN)
        return d

    def test_request_password_reset_token(self):
        """Test that the request_password_reset_token method works ok."""
        d = Deferred()
        processor = self.create_mock_processor()
        processor.request_password_reset_token(EMAIL)
        self.patch(ubuntu_sso.main, "blocking", self.fake_ok_blocking)
        self.mocker.replay()

        def verify(result):
            self.assertEqual(result, EMAIL)
            d.callback(result)

        client = SSOLogin(self.mockbusname,
                          sso_login_processor_class=self.mockprocessorclass)
        self.patch(client, "PasswordResetTokenSent", verify)
        self.patch(client, "PasswordResetError", d.errback)
        client.request_password_reset_token(EMAIL)
        return d

    def test_request_password_reset_token_error(self):
        """Test the request_password_reset_token method fails as expected."""
        d = Deferred()

        self.create_mock_processor().request_password_reset_token(EMAIL)
        self.mocker.result(True)
        self.patch(ubuntu_sso.main, "blocking", self.fake_err_blocking)
        self.mocker.replay()

        def verify(error):
            self.assertIsInstance(error, BlockingSampleException)
            d.callback("Ok")

        client = SSOLogin(self.mockbusname,
                          sso_login_processor_class=self.mockprocessorclass)
        self.patch(client, "PasswordResetTokenSent", d.errback)
        self.patch(client, "PasswordResetError", verify)
        client.request_password_reset_token(EMAIL)
        return d

    def test_set_new_password(self):
        """Test that the set_new_password method works ok."""
        d = Deferred()
        self.create_mock_processor().set_new_password(EMAIL, EMAIL_TOKEN,
                                                      PASSWORD)
        self.mocker.result(EMAIL)
        self.patch(ubuntu_sso.main, "blocking", self.fake_ok_blocking)
        self.mocker.replay()

        def verify(result):
            self.assertEqual(result, EMAIL)
            d.callback(result)

        client = SSOLogin(self.mockbusname,
                          sso_login_processor_class=self.mockprocessorclass)
        self.patch(client, "PasswordChanged", verify)
        self.patch(client, "PasswordChangeError", d.errback)
        client.set_new_password(EMAIL, EMAIL_TOKEN, PASSWORD)
        return d

    def test_set_new_password_error(self):
        """Test that the set_new_password method fails as expected."""
        d = Deferred()
        expected_result = "expected result"

        self.create_mock_processor().set_new_password(EMAIL, EMAIL_TOKEN,
                                                      PASSWORD)
        self.mocker.result(expected_result)
        self.patch(ubuntu_sso.main, "blocking", self.fake_err_blocking)
        self.mocker.replay()

        def verify(error):
            self.assertIsInstance(error, BlockingSampleException)
            d.callback("Ok")

        client = SSOLogin(self.mockbusname,
                          sso_login_processor_class=self.mockprocessorclass)
        self.patch(client, "PasswordChanged", d.errback)
        self.patch(client, "PasswordChangeError", verify)
        client.set_new_password(EMAIL, EMAIL_TOKEN, PASSWORD)
        return d


class BlockingFunctionTestCase(TestCase):
    """Tests for the "blocking" function."""

    def test_blocking(self):
        """Test the normal behaviour."""
        d = Deferred()
        expected_result = "expected result"

        def f():
            return expected_result

        def verify(result):
            self.assertEqual(result, expected_result)
            d.callback(result)

        blocking(f, verify, d.errback)
        return d

    def test_blocking_error(self):
        """Test the behaviour when an Exception is raised."""
        d = Deferred()
        expected_error_message = "expected error message"

        def f():
            raise BlockingSampleException(expected_error_message)

        def verify(error):
            self.assertEqual(error, expected_error_message)
            d.callback("Ok")

        blocking(f, d.errback, verify)
        return d


class KeyringCredentialsTestCase(MockerTestCase):
    """Checks the functions that access the keyring."""

    def test_keyring_store_cred(self):
        """Verify the method that stores credentials."""
        token_value = TOKEN
        mockKeyringClass = self.mocker.replace("ubuntu_sso.keyring.Keyring")
        mockKeyringClass(APP_NAME)
        mockKeyring = self.mocker.mock()
        self.mocker.result(mockKeyring)
        mockKeyring.set_ubuntusso_attr(token_value)
        self.mocker.replay()

        keyring_store_credentials(APP_NAME, token_value)

    def test_keyring_get_cred(self):
        """The method returns the right token."""
        mockKeyringClass = self.mocker.replace("ubuntu_sso.keyring.Keyring")
        mockKeyringClass(APP_NAME)
        mockKeyring = self.mocker.mock()
        self.mocker.result(mockKeyring)
        mockKeyring.get_ubuntusso_attr()
        self.mocker.result(TOKEN)
        self.mocker.replay()

        token = keyring_get_credentials(APP_NAME)
        self.assertEqual(token, TOKEN)

    def test_keyring_get_cred_not_found(self):
        """The method returns None when the token is not found."""
        mockKeyringClass = self.mocker.replace("ubuntu_sso.keyring.Keyring")
        mockKeyringClass(APP_NAME)
        mockKeyring = self.mocker.mock()
        self.mocker.result(mockKeyring)
        mockKeyring.get_ubuntusso_attr()
        self.mocker.result(None)
        self.mocker.replay()

        token = keyring_get_credentials(APP_NAME)
        self.assertEqual(token, None)

    def test_keyring_get_old_cred_found(self):
        """The method returns a new set of creds if old creds are found."""
        sample_oauth_token = "sample oauth token"
        sample_oauth_secret = "sample oauth secret"
        old_creds = {
            "oauth_token": sample_oauth_token,
            "oauth_token_secret": sample_oauth_secret,
        }

        mockKeyringClass = self.mocker.replace("ubuntu_sso.keyring.Keyring")
        mockKeyringClass(U1_APP_NAME)
        mockKeyring = self.mocker.mock()
        self.mocker.result(mockKeyring)
        mockKeyring.get_ubuntusso_attr()
        self.mocker.result(None)

        mockKeyringClass(OLD_KEY_NAME)
        mockKeyring2 = self.mocker.mock()
        self.mocker.result(mockKeyring2)
        mockKeyring2.get_ubuntusso_attr()
        self.mocker.result(old_creds)

        self.mocker.replay()

        new_creds = keyring_get_credentials(U1_APP_NAME)
        self.assertIn("token", new_creds)
        self.assertEqual(new_creds["token"], sample_oauth_token)
        self.assertIn("token_secret", new_creds)
        self.assertEqual(new_creds["token_secret"], sample_oauth_secret)

    def test_keyring_get_old_cred_not_found(self):
        """The method returns None if no old nor new credentials found."""
        mockKeyringClass = self.mocker.replace("ubuntu_sso.keyring.Keyring")
        mockKeyringClass(U1_APP_NAME)
        mockKeyring = self.mocker.mock()
        self.mocker.result(mockKeyring)
        mockKeyring.get_ubuntusso_attr()
        self.mocker.result(None)

        mockKeyringClass(OLD_KEY_NAME)
        mockKeyring2 = self.mocker.mock()
        self.mocker.result(mockKeyring2)
        mockKeyring2.get_ubuntusso_attr()
        self.mocker.result(None)

        self.mocker.replay()

        token = keyring_get_credentials(U1_APP_NAME)
        self.assertEqual(token, None)


class RegisterSampleException(Exception):
    """A mock exception thrown just when testing."""


class CredentialsTestCase(TestCase, MockerTestCase):
    """Tests for the credentials related DBus methods."""

    timeout = 5

    def test_find_credentials(self):
        """find_credentials immediately returns the token when found."""
        expected_token = "expected token"
        kgt = self.mocker.replace("ubuntu_sso.main.keyring_get_credentials")
        kgt(ARGS)
        self.mocker.result(expected_token)
        self.mocker.replay()

        client = SSOCredentials(self.mocker.mock())
        token = client.find_credentials(APP_NAME)
        self.assertEqual(token, expected_token)

    def test_credentials_not_found(self):
        """find_credentials immediately returns '' when no token found."""
        expected_creds = {}
        kgt = self.mocker.replace("ubuntu_sso.main.keyring_get_credentials")
        kgt(ARGS)
        self.mocker.result(None)
        self.mocker.replay()

        client = SSOCredentials(self.mocker.mock())
        token = client.find_credentials(APP_NAME)
        self.assertEqual(token, expected_creds)

    def test_login_or_register(self):
        """login_or_register_... throws the signal when token is found."""
        expected_creds = TOKEN
        d = Deferred()

        def verify(result):
            self.assertEqual(result, expected_creds)
            d.callback("ok")

        kgt = self.mocker.replace("ubuntu_sso.main.keyring_get_credentials")
        kgt(ARGS)
        self.mocker.result(expected_creds)
        self.mocker.replay()
        client = SSOCredentials(self.mocker.mock())
        self.patch(client, "_show_login_or_register_ui", self.fail)
        self.patch(client, "CredentialsFound", verify)
        self.patch(client, "CredentialsError", self.fail)
        client.login_or_register_to_get_credentials(*LOGIN_OR_REGISTER_ARGS)
        return d

    def test_login_or_register_not_found(self):
        """Check that login_or_register_... opens the ui when no cred found."""
        d = Deferred()

        def verify(result, *a):
            self.assertEqual(result, APP_NAME)
            d.callback("ok")

        kgt = self.mocker.replace("ubuntu_sso.main.keyring_get_credentials")
        kgt(ARGS)
        self.mocker.result(None)
        self.mocker.replay()

        client = SSOCredentials(self.mocker.mock())
        self.patch(client, "_show_login_or_register_ui", verify)
        self.patch(client, "CredentialsFound", self.fail)
        self.patch(client, "CredentialsError", self.fail)
        client.login_or_register_to_get_credentials(*LOGIN_OR_REGISTER_ARGS)
        return d

    def test_login_or_register_problem(self):
        """login_or_register_... returns the right signal on error."""
        expected_error = "Sample Error - not for resale"
        d = Deferred()

        def verify(result, *a):
            self.assertEqual(result, expected_error)
            d.callback("ok")

        kgt = self.mocker.replace("ubuntu_sso.main.keyring_get_credentials")
        kgt(ARGS)
        self.mocker.throw(RegisterSampleException(expected_error))
        self.mocker.replay()

        client = SSOCredentials(self.mocker.mock())
        self.patch(client, "_show_login_or_register_ui", self.fail)
        self.patch(client, "CredentialsFound", self.fail)
        self.patch(client, "CredentialsError", verify)
        client.login_or_register_to_get_credentials(*LOGIN_OR_REGISTER_ARGS)
        return d

    def test_login_only(self):
        """login_only_... throws the signal when token is found."""
        expected_creds = TOKEN
        d = Deferred()

        def verify(result):
            self.assertEqual(result, expected_creds)
            d.callback("ok")

        kgt = self.mocker.replace("ubuntu_sso.main.keyring_get_credentials")
        kgt(ARGS)
        self.mocker.result(expected_creds)
        self.mocker.replay()
        client = SSOCredentials(self.mocker.mock())
        self.patch(client, "_show_login_only_ui", self.fail)
        self.patch(client, "CredentialsFound", verify)
        self.patch(client, "CredentialsError", self.fail)
        client.login_to_get_credentials(*LOGIN_ONLY_ARGS)
        return d

    def test_login_only_not_found(self):
        """Check that login_only_... opens the ui when no cred found."""
        d = Deferred()

        def verify(result, *a):
            self.assertEqual(result, APP_NAME)
            d.callback("ok")

        kgt = self.mocker.replace("ubuntu_sso.main.keyring_get_credentials")
        kgt(ARGS)
        self.mocker.result(None)
        self.mocker.replay()

        client = SSOCredentials(self.mocker.mock())
        self.patch(client, "_show_login_only_ui", verify)
        self.patch(client, "CredentialsFound", self.fail)
        self.patch(client, "CredentialsError", self.fail)
        client.login_to_get_credentials(*LOGIN_ONLY_ARGS)
        return d

    def test_login_only_problem(self):
        """login_only_... returns the right signal on error."""
        expected_error = "Sample Error - not for resale"
        d = Deferred()

        def verify(result, *a):
            self.assertEqual(result, expected_error)
            d.callback("ok")

        kgt = self.mocker.replace("ubuntu_sso.main.keyring_get_credentials")
        kgt(ARGS)
        self.mocker.throw(RegisterSampleException(expected_error))
        self.mocker.replay()

        client = SSOCredentials(self.mocker.mock())
        self.patch(client, "_show_login_only_ui", self.fail)
        self.patch(client, "CredentialsFound", self.fail)
        self.patch(client, "CredentialsError", verify)
        client.login_to_get_credentials(*LOGIN_ONLY_ARGS)
        return d


class CredentialsGUITestCase(TestCase):
    """login_or_register_to_get_credentials opens the proper GUI."""

    def setUp(self):
        """Init."""
        idle_add = lambda f, *args, **kwargs: f(*args, **kwargs)
        self.patch(gobject, "idle_add", idle_add)
        self.patch(ubuntu_sso.main, 'keyring_get_credentials', lambda n: None)

        self.memento = MementoHandler()
        self.memento.setLevel(logging.DEBUG)
        logger.addHandler(self.memento)

        self.client = SSOCredentials(None)
        self.login_or_register = \
            self.client.login_or_register_to_get_credentials
        self.login_only = self.client.login_to_get_credentials

    def test_login_or_register_opens_proper_ui(self):
        """The proper GUI is created if the token is None."""

        self.args = None
        self.kwargs = None

        class FakedUbuntuSSOClientGUI(object):
            """Fake a SSO GUI."""

            def __init__(sself, *args, **kwargs):
                self.args = args
                self.kwargs = kwargs
                sself.connect = lambda *a: None

        self.patch(gui, "UbuntuSSOClientGUI", FakedUbuntuSSOClientGUI)
        self.login_or_register(*LOGIN_OR_REGISTER_ARGS)

        self.assertEqual(self.args, LOGIN_OR_REGISTER_GUI_ARGS)
        self.assertEqual(self.kwargs, {})

    def test_ui_exceptions_are_logged(self):
        """If GUI fails the exception is logged."""

        self.expected = None

        def fail(*args, **kwargs):
            self.expected = AssertionError((args, kwargs))
            raise self.expected

        self.patch(gui, "UbuntuSSOClientGUI", fail)
        self.login_or_register(*LOGIN_OR_REGISTER_ARGS)

        self.assertTrue(1, len(self.memento.records))
        self.assertTrue(self.memento.check(logging.ERROR,
                                           LOGIN_OR_REGISTER_GUI_ARGS))
        exc_info = self.memento.records[0].exc_info
        self.assertIn(self.expected, exc_info)

    def test_ui_signals_are_connected(self):
        """Signals from GUI are properly connected."""

        self.signals = []

        class FakedUbuntuSSOClientGUI(object):
            """Fake a SSO GUI."""

            def __init__(sself, *args, **kwargs):
                sself.connect = lambda *a: self.signals.append(a)

        expected = [
            (gui.SIG_LOGIN_SUCCEEDED, self.client._login_success_cb),
            (gui.SIG_LOGIN_FAILED, self.client._login_error_cb),
            (gui.SIG_REGISTRATION_SUCCEEDED, self.client._login_success_cb),
            (gui.SIG_REGISTRATION_FAILED, self.client._login_error_cb),
            (gui.SIG_USER_CANCELATION, self.client._login_auth_denied_cb),
        ]
        self.patch(gui, "UbuntuSSOClientGUI", FakedUbuntuSSOClientGUI)
        self.login_or_register(*LOGIN_OR_REGISTER_ARGS)
        self.assertEqual(self.signals, expected)

    def test_login_only_opens_proper_ui(self):
        """The proper GUI is created if the token is None."""

        self.args = None
        self.kwargs = None

        class FakedUbuntuSSOClientGUI(object):
            """Fake a SSO GUI."""

            def __init__(sself, *args, **kwargs):
                self.args = args
                self.kwargs = kwargs
                sself.connect = lambda *a: None

        self.patch(gui, "UbuntuSSOClientGUI", FakedUbuntuSSOClientGUI)
        self.login_only(*LOGIN_ONLY_ARGS)

        self.assertEqual(self.args, LOGIN_ONLY_GUI_ARGS)
        self.assertEqual(self.kwargs, {})

    def test_login_only_ui_exceptions_are_logged(self):
        """If GUI fails the exception is logged."""

        self.expected = None

        def fail(*args, **kwargs):
            self.expected = AssertionError((args, kwargs))
            raise self.expected

        self.patch(gui, "UbuntuSSOClientGUI", fail)
        self.login_only(*LOGIN_ONLY_ARGS)

        self.assertTrue(1, len(self.memento.records))
        self.assertTrue(self.memento.check(logging.ERROR,
                                           LOGIN_ONLY_GUI_ARGS))
        exc_info = self.memento.records[0].exc_info
        self.assertIn(self.expected, exc_info)

    def test_login_only_ui_signals_are_connected(self):
        """Signals from GUI are properly connected."""

        self.signals = []

        class FakedUbuntuSSOClientGUI(object):
            """Fake a SSO GUI."""

            def __init__(sself, *args, **kwargs):
                sself.connect = lambda *a: self.signals.append(a)

        expected = [
            (gui.SIG_LOGIN_SUCCEEDED, self.client._login_success_cb),
            (gui.SIG_LOGIN_FAILED, self.client._login_error_cb),
            (gui.SIG_REGISTRATION_SUCCEEDED, self.client._login_success_cb),
            (gui.SIG_REGISTRATION_FAILED, self.client._login_error_cb),
            (gui.SIG_USER_CANCELATION, self.client._login_auth_denied_cb),
        ]
        self.patch(gui, "UbuntuSSOClientGUI", FakedUbuntuSSOClientGUI)
        self.login_only(*LOGIN_ONLY_ARGS)
        self.assertEqual(self.signals, expected)
